from typing import Optional
from typing import Tuple

from ._common_imports import classad
from .htcondor2_impl import _send_generic_payload_command
from ._schedd import Schedd

import enum

class DagmanConnectionError(Exception):
    """
    Custom error for internal use to be raised when a DAGMan
    connection fails and should be retried one more time.

    :meta private:
    """
    def __init__(self, msg: Optional[str] = None) -> None:
        self.message = msg if msg is not None else "Failed to connect to DAGMan"
        super().__init__(self.message)


class DAGMan():
    """
    The DAGMan client that acts as a connection to a running DAGMan
    process for sending commands.
    """

    # Actual DaemonCore command int from condor_commands.h
    _CMD_DAGMAN_GENERIC = 61500

    # Success return code of low level api calls and internal methods
    _SUCCESS = 0

    # Keep consistent with dagman_commands.h enumeration
    class Command(enum.IntEnum):
        """
        Internal mapping of specific generic payload DAGMan commands.

        :meta private:
        """

        HALT = 1
        RESUME = 2

    class Failure(enum.IntEnum):
        """
        Error Codes returned from lower level api calls and internal
        methods.

        :meta private:
        """

        #### Return values from _send_generic_payload
        OPEN_SOCKET = 1        # Failed to open command socker
        SEND_PAYLOAD = 2       # Failed to send request payload to DAGMan
        GET_RESULT = 3         # Failed to get return payload form DAGMan
        #### Internal method failure codes
        NO_ID = 4              # No DAG id set when calling __locate()
        CONNECT_SCHEDD = 5     # Failed to connect with Schedd to get DAGMan contact info
        GET_CONTACT = 6        # Failed to get DAGMan contact info (Schedd failure result ad)
        DENIED = 7             # Failed to get DAGMan contact info (Specifically not permitted)

    class ContactInfo():
        """
        DAGMan contact information. Sinful address to send command
        and secret DAGMan will use to verify trust.

        :meta private:
        """
        def __init__(self, addr: str, secret: str) -> None:
            self.addr = addr
            self.secret = secret


    def __init__(self, dag_id: int) -> None:
        """
        :param dag_id: The :ad-attr:`ClusterId` of a DAGMan job to
                       locate once the first command is issued.
        """
        self._contact = None

        if not isinstance(dag_id, int):
            raise TypeError("dag_id must be an integer")
        self._dag_id = dag_id


    @property
    def dag_id(self) -> int:
        """
        The :ad-attr:`ClusterId` of a DAGMan job this object will send commands.
        """
        return self._dag_id


    # Note: Pause is not a public option
    def halt(self, reason: Optional[str] = None, pause: Optional[bool] = None) -> Tuple[bool, str]:
        """halt(reason = None) -> tuple[bool, str]
        Inform DAGMan to halt a DAGs progress.

        :param reason: A message for why the DAG was halted to be
                       printed in the DAGs debug log.
        :return: Command success and result message.
        """
        request = classad.ClassAd({"DagCommand" : self.Command.HALT})

        if pause is not None:
            request["IsPause"] = pause

        if reason is not None:
            request["HaltReason"] = reason

        result, ret, err = self.__contact_dagman(self._CMD_DAGMAN_GENERIC, request)

        if ret != self._SUCCESS:
            return (False, err)

        if not result.get("Result", False):
            return (False, result.get('ErrorString', 'Unknown'))

        action = "Paused" if pause is not None and pause else "Halted"

        return (True, f"{action} DAG {self.dag_id}")


    def resume(self) -> Tuple[bool, str]:
        """
        Inform DAGMan to resume a hatled DAGs progress.

        :return: Command success and result message.
        """
        request = classad.ClassAd({"DagCommand" : self.Command.RESUME})

        result, ret, err = self.__contact_dagman(self._CMD_DAGMAN_GENERIC, request)

        if ret != self._SUCCESS:
            return (False, err)

        if not result.get("Result", False):
            return (False, result.get('ErrorString', 'Unknown'))

        return (True, f"Resumed DAG {self.dag_id}")


    def __contact_dagman(self, cmd: int, request: classad.ClassAd) -> Tuple[classad.ClassAd, int, Optional[str]]:
        """
        Attempt to send command with payload to DAGMan and recieve a result ClassAd.

        :param cmd: DAGMan command integer.
        :param request: The :class:`classad.ClassAd` containing the specific
                        commands information.

        :return: Result ClassAd, error code, and failure reason.

        :meta private:
        """

        # Get the DAGMan contact information if we don't already have it
        if self._contact is None:
            ret, err = self.__locate()
            if ret != self._SUCCESS:
                return (None, ret, err)

        # Add the contact secret to the request ad so DAGMan can verify us
        request["Secret"] = self._contact.secret

        # Attempt to send command to DAGMan
        try:
            result, ret, err = _send_generic_payload_command(self._contact.addr, int(cmd), request._handle)

            # Specific errors (initial connection and not trusted) may be a result of DAGMan going away
            if ret in [self.Failure.OPEN_SOCKET, self.Failure.SEND_PAYLOAD]:
                raise DagmanConnectionError(f"Connection to {self._contact.addr} ({self.dag_id}) failed: {err}")
            elif not result.get("Trusted", False):
                raise DagmanConnectionError(f"Connection to {self._contact.addr} ({self.dag_id}) failed: {result.get('ErrorString')}")
        except DagmanConnectionError as e:
            # Reset DAGMan contact information and retry command
            ret, err = self.__locate()
            if ret != self._SUCCESS:
                return (None, ret, err)

            result, ret, err = _send_generic_payload_command(self._contact.addr, int(cmd), request._handle)

        return (result, ret, err)


    def __locate(self) -> Tuple[int, str]:
        """
        Query the local Schedd for DAGMan contact information.

        :return: Error code and failure reason.

        :meta private:
        """

        # Reset contact information
        self._contact = None

        if self.dag_id is None:
            return (self.Failure.NO_ID, "No DAG ID specified")

        try:
            schedd = Schedd()
            ad = schedd._get_dag_contact_info(self.dag_id)
        except Exception as e:
            return (self.Failure.CONNECT_SCHEDD, f"Failed to get contact information: {str(e)}")

        if not ad.get("Result", False):
            ret = self.Failure.DENIED if result.get("Permitted", False) else self.Failure.GET_CONTACT
            return (ret, ad.get("ErrorString", "Unknown"))

        self._contact = DAGMan.ContactInfo(ad["Address"], ad["Secret"])
        return (self._SUCCESS, None)




#!/usr/bin/env pytest

# Test explicit use of an in-memory token in the bindings

import htcondor2 as htcondor
import time
import logging

from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@standup
def condor(test_dir):
    with Condor(
        local_dir=test_dir / "condor",
        config={
            "SEC_DEFAULT_AUTHENTICATION_METHODS": "TOKEN,FS",
            "DAEMON_LIST": "COLLECTOR MASTER",
            "ALLOW_WRITE": "*",
            "ENABLE_IPV6": "false",
            "COLLECTOR_DEBUG": "D_SECURITY:2",
        }
    ) as condor:
        yield condor

@action
def bespoke_token(condor):
    rv = condor.run_command(
        ["condor_token_create", "-identity", "vulture@foo"],
        timeout=8,
    )
    assert rv.returncode == 0
    return rv.stdout


class TestPythingBindingsToken:

    def test_ping_with_token(self, condor, bespoke_token):
        # We need condor.use_config() so that the locate() of a local daemon
        # looks in the correct LOG directory for the daemon address file.
        with condor.use_config():
            collector = condor.get_local_collector()
            collector_ad = None
            collector_ad = collector.locate(htcondor.DaemonType.Collector)

        ctxt = htcondor.SecurityContext(token=bespoke_token)
        ping_ad = htcondor.ping(collector_ad, "WRITE", ctxt)

        assert ping_ad["Authentication"] == "YES"
        assert ping_ad["AuthMethods"] == "IDTOKENS"
        assert ping_ad["MyRemoteUserName"] == "vulture@foo"



#!/usr/bin/env pytest

#testreq: personal
"""<<CONDOR_TESTREQ_CONFIG
	# make sure that file transfer plugins are enabled (might be disabled by default)
	ENABLE_URL_TRANSFERS = true
	FILETRANSFER_PLUGINS = $(LIBEXEC)/curl_plugin $(LIBEXEC)/data_plugin
"""
#endtestreq


#Job Ad attribute: TransferInputStats regression test
#
#The attribute 'TransferInputStats' nested attributes were
#overflowing due to big numbers being shoved into int variables.
#This test is to make sure that large files transferred don't
#result in the nested attributes have byte transfer sizes less
#than 0.
#


from ornithology import *
from time import sleep

#-----------------------------------------------------------------------------------------
@action
def job_python_file(test_dir):
     job_python_file = test_dir / "xferstats.py"
     contents = format_script(
     """#!/usr/bin/env python3

import classad2 as classad
import json
import os
import posixpath
import shutil
import socket
import sys
import time
from glob import glob as find

from urllib.parse import urlparse

DEFAULT_TIMEOUT = 30
PLUGIN_VERSION = '1.0.0'

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_AUTHENTICATION_REFRESH = 2


def print_help(stream = sys.stderr):
    help_msg = '''Usage: {0} -infile <input-filename> -outfile <output-filename>
       {0} -classad

Options:
  -classad                    Print a ClassAd containing the capablities of this
                              file transfer plugin.
  -infile <input-filename>    Input ClassAd file
  -outfile <output-filename>  Output ClassAd file
  -upload                     Indicates this transfer is an upload (default is
                              download)
'''
    stream.write(help_msg.format(sys.argv[0]))

def print_capabilities():
    capabilities = {
         'MultipleFileSupport': True,
         'PluginType': 'FileTransfer',
         # SupportedMethods indicates which URL methods/types this plugin supports
         'SupportedMethods': 'xferstats',
         'Version': PLUGIN_VERSION,
    }
    sys.stdout.write(classad.ClassAd(capabilities).printOld())

def parse_args():

    # The only argument lists that are acceptable are
    # <this> -classad
    # <this> -infile <input-filename> -outfile <output-filename>
    # <this> -outfile <output-filename> -infile <input-filename>
    if not len(sys.argv) in [2, 5, 6]:
        print_help()
        sys.exit(EXIT_FAILURE)

    # If -classad, print the capabilities of the plugin and exit early
    if (len(sys.argv) == 2) and (sys.argv[1] == '-classad'):
        print_capabilities()
        sys.exit(EXIT_SUCCESS)

    # If -upload, set is_upload to True and remove it from the args list
    is_upload = False
    if '-upload' in sys.argv[1:]:
        is_upload = True
        sys.argv.remove('-upload')

    # -infile and -outfile must be in the first and third position
    if not (
            ('-infile' in sys.argv[1:]) and
            ('-outfile' in sys.argv[1:]) and
            (sys.argv[1] in ['-infile', '-outfile']) and
            (sys.argv[3] in ['-infile', '-outfile']) and
            (len(sys.argv) == 5)):
        print_help()
        sys.exit(1)
    infile = None
    outfile = None
    try:
        for i, arg in enumerate(sys.argv):
            if i == 0:
                continue
            elif arg == '-infile':
                infile = sys.argv[i+1]
            elif arg == '-outfile':
                outfile = sys.argv[i+1]
    except IndexError:
        print_help()
        sys.exit(EXIT_FAILURE)

    return {'infile': infile, 'outfile': outfile, 'upload': is_upload}

def format_error(error):
    return '{0}: {1}'.format(type(error).__name__, str(error))

def get_error_dict(error, url = ''):
    error_string = format_error(error)
    error_dict = {
        'TransferSuccess': False,
        'TransferError': error_string,
        'TransferUrl': url,
    }

    return error_dict

class XferStatsPlugin:

    # Extract whatever information we want from the url provided.
    def Xfile_path(self):
        # Get test dir path and remove the script name from the end
        path = os.path.realpath(__file__)
        # Split path and end filename
        parts = path.rsplit("/",1)
        # Return path / xferIn.txt
        return (parts[0] + "/xferIn.txt")

    def download_file(self, url, local_file_path):

        start_time = time.time()

        file_size = 0
        test = "-"
        f = open(self.Xfile_path(), "r")
        test = f.readline().strip()
        file_size = int(test[:-1]) * 1024 * 1024 * 1024
        f.close()
        print(f'\\nTest-: {test} file')
        print(f'Size-: {file_size} bytes')

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': 'xferstats',
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }

        return transfer_stats

    def upload_file(self, url, local_file_path):

        start_time = time.time()

        file_size = 0
        test = "-"
        f = open(self.Xfile_path(), "r")
        test = f.readline().strip()
        file_size = int(test[:-1]) * 1024 * 1024 * 1024
        f.close()
        print(f'\\nTest-: {test} file')
        print(f'Size-: {file_size} bytes')

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': 'xferstats',
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }

        return transfer_stats


if __name__ == '__main__':

    # Start by parsing input arguments
    try:
        args = parse_args()
    except Exception:
        sys.exit(EXIT_FAILURE)

    xferstats_plugin = XferStatsPlugin()

    # Parse in the classads stored in the input file. 
    # Each ad represents a single file to be transferred.
    try:
        infile_ads = classad.parseAds(open(args['infile'], 'r'))
    except Exception as err:
        try:
            with open(args['outfile'], 'w') as outfile:
                outfile_dict = get_error_dict(err)
                outfile.write(str(classad.ClassAd(outfile_dict)))
        except Exception:
            pass
        sys.exit(EXIT_FAILURE)

    # Now iterate over the list of classads and perform the transfers.
    try:
        with open(args['outfile'], 'w') as outfile:
            for ad in infile_ads:
                try:
                    if not args['upload']:
                        outfile_dict = xferstats_plugin.download_file(ad['Url'], ad['LocalFileName'])
                    else:
                        outfile_dict = xferstats_plugin.upload_file(ad['Url'], ad['LocalFileName'])

                    outfile.write(str(classad.ClassAd(outfile_dict)))

                except Exception as err:
                    try:
                        outfile_dict = get_error_dict(err, url = ad['Url'])
                        outfile.write(str(classad.ClassAd(outfile_dict)))
                    except Exception:
                        pass
                    sys.exit(EXIT_FAILURE)

    except Exception:
        sys.exit(EXIT_FAILURE)
     """)
     write_file(job_python_file, contents)
     return job_python_file

#-----------------------------------------------------------------------------------------
@action
def job_shell_file(test_dir, job_python_file):
    job_shell_file = test_dir / "xferstats.sh"
    contents = format_script(
    f"""
        #!/bin/bash
        exec {job_python_file} $@ >> {test_dir}/plugin.log 2>&1
    """
    )
    write_file(job_shell_file, contents)
    return job_shell_file

#-----------------------------------------------------------------------------------------
#Fixture to run a simple job with varying sized transfer files
@action(params={
"0kb File":"0K",
"2gb File":"2G",
"4gb File":"4G",
"8gb File":"8G",
"16gb File":"16G",
})
def run_file_xfer_job(default_condor,test_dir,request, job_shell_file, path_to_sleep):
     #Write Test File size to input file
     xfer_file = "xferIn.txt"
     _file = open(xfer_file,"w")
     _file.write(request.param)
     _file.close()
     #Submit a list command job with output so we can varify file sizes
     job_handle = default_condor.submit(
          {
               "executable":path_to_sleep,
               "arguments":"0",
               "should_transfer_files":"Yes",
               "transfer_input_files":f"xferstats://{test_dir}/{xfer_file}",
               "transfer_plugins":f"xferstats={job_shell_file}",
               "log":(test_dir / "{0}-test.log".format(request.param)).as_posix(),
          },
          count=1
     )
     #Wait for job to complete. Gave a hefty amount of time due to large files being transfered
     clustId = job_handle.clusterid
     assert job_handle.wait(condition=ClusterState.all_complete,timeout=30)
     schedd = default_condor.get_local_schedd()
     #once job is done get job ad attributes needed for test verification
     job_ad = []
     for i in range(0,5):
          hist_itr = schedd.history(
               constraint=f"ClusterId=={clustId}",
               projection=["TransferInput","TransferInputStats","TransferOutputStats"],
               match=1,
          )
          job_ad = list(hist_itr)
          if len(job_ad) > 0:
              break
          sleep(1)
     assert len(job_ad) > 0
     #Return job ad for checking attributes
     return (request.param,job_ad)

#=========================================================================================
#JobSubmitMethod Tests
class TestTransferStatsOverflowRegression:

     #Test that a job submited with value < Minimum doesn't add JobSubmitMethod attr
     def test_transfer_stats_bytes_dont_overflow(self,run_file_xfer_job):
          passed = True
          is_0K = False
          count = 0
          #For every classad returned (should only be 1) check statistics
          for ad in run_file_xfer_job[1]:
               count += 1
               if "0K" in run_file_xfer_job[0]:
                    is_0K = True
               #For every statistic in TransferInputStats
               for stat in ad["TransferInputStats"]:
                    #Check if stat is a 'SizeBytes' statistic
                    if "SizeBytes" in stat:
                         #Get value
                         value = ad["TransferInputStats"][stat]
                         #If value is less than 0 overflow occured. Output error and fail test
                         if value < 0:
                              passed = False
                              print(f"Error: {stat} value is {value}. The value should be 0 or greater.")
                         #If it isn't the 0Kb file and value is 0 then overflow occured. Output error and fail test.
                         elif is_0K is False and value == 0:
                              passed = False
                              print(f"Error: {stat} value is {value} for a file larger than 0 bytes. The value should not be 0.")
               #For every statistic in TransferOutputStats
               for stat in ad["TransferOutputStats"]:
                    #Check if stat is a 'SizeBytes' statistic
                    if "SizeBytes" in stat:
                         #Get value
                         value = ad["TransferOutputStats"][stat]
                         #If value is less than 0 overflow occured. Output error and fail test
                         if value < 0:
                              passed = False
                              print(f"Error: {stat} value is {value}. The value should be 0 or greater.")
                         #If it isn't the 0Kb file and value is 0 then overflow occured. Output error and fail test.
                         elif is_0K is False and value == 0:
                              passed = False
                              print(f"Error: {stat} value is {value} for a file larger than 0 bytes. The value should not be 0.")
          if count != 1:
              passed = False
              print(f"Error: history() returned {count} job ads. Expected 1 ad.")
          #Assert Pass or Fail here
          assert passed





#!/usr/bin/env pytest

from ornithology import (
    action,
    Condor
)

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

import time


@action
def the_condor(test_dir):
    local_dir = test_dir / "condor"

    with Condor(
        local_dir=local_dir,
        config={
            "TPIPEH":       "$(LIBEXEC)/test_std_pipe_handlerd",
            "DAEMON_LIST":  "$(DAEMON_LIST) TPIPEH",
            # Ornithology needs this, for some reason.
            "TPIPEH_LOG":    f"$(LOG)/TestStdPipeHandlerDLog",
        },
    ) as condor:
        yield condor


@action
def the_daemon_log(the_condor):
    the_log = the_condor._get_daemon_log('TPIPEH')
    the_open_log = the_log.open()

    for timeout in range(1, 30):
        for line in the_open_log.read():
            print( f"  >>> {line}" )
            if "EXITING WITH STATUS" in line.message:
                return the_log
        time.sleep(2)
    assert False


class TestStdPipeHandler:

    def test_std_pipe_handler(self, the_condor, the_daemon_log):
        # I'm not at all sure I can just re-open the daemon log.
        the_log = the_condor._get_daemon_log('TPIPEH')

        found_complete_read = False
        for line in the_log.open().read():
            print( f"  >>> {line}" )
            if not found_complete_read:
                if "Read 0x7 0x4 0x3 0x71 0x22 0x70 0x61 0x55 0x24 0xaa 0xbc 0xad 0xff 0xee 0xcb 0x7 0x4 0x3 0x71 0x22 so far." in line.message:
                    found_complete_read = True
                continue
            else:
                if "Done!" in line.message:
                    return
                if "Read " in line.message and "so far." in line.message:
                    assert False
        assert False



#!/usr/bin/env pytest

import pytest
from pathlib import Path

import logging
import htcondor2

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


# Confirm that we get the same itemdata from using setQArgs() that
# we do from specifying the qargs in the call to itemdata().  This
# assumes that the former is correct, but _does_ catch a potential
# problem in the implementation of the latter, where submit macros
# in the qargs weren't expanded (because they were unset).


submit_template = {
    "executable": "/bin/sleep",
    "args": "60",
    "MY.my_var": "$(my_var)",
    "submit_var": "svr",
}


@pytest.fixture
def the_list(test_dir):
    list_path = test_dir / "list.txt"
    list_path.write_text('''"a"
"1"
"2"
"iii"
''')
    return str(list_path)


@pytest.fixture
def from_setqargs(the_list):
    submit_object = htcondor2.Submit(submit_template)
    submit_object.setQArgs(f"2 $(submit_var) FROM {the_list}")
    itemdata = submit_object.itemdata()
    return list(itemdata)


@pytest.fixture
def from_itemdata(the_list):
    submit_object = htcondor2.Submit(submit_template)
    itemdata = submit_object.itemdata(qargs=f"2 $(submit_var) FROM {the_list}")
    return list(itemdata)


class TestSubmitItemdataArgs():

    def test_identical_to_setqargs(self, from_setqargs, from_itemdata):
        assert from_setqargs == from_itemdata



#!/usr/bin/env pytest

import pytest

import htcondor2


# The block between the lines is a speculative/test implementation and not at
# all necessary for this test.
# -----------------------------------------------------------------------------

TEST_JOBS = {}

def jobs(*args, params):
    # Record the name and contents of each test case.
    global TEST_JOBS
    TEST_JOBS.update(params)


    def decorator(func):
        return pytest.fixture(
            scope="class",
            ids=params.keys(),
            params=params.keys(),
        )(func)

    if len(args) == 1:
        return decorator(args[0])

    return decorator


TEST_JOB_HANDLES = None
def make_test_job(the_condor, test_job_ID):
    global TEST_JOB_HANDLES

    if TEST_JOB_HANDLES is None:
        TEST_JOB_HANDLES = {}
        schedd = the_condor.get_local_schedd()
        for id, description in TEST_JOBS.items():
            submit = htcondor2.Submit(description)
            handle = schedd.submit(
                description=submit,
                count=1,
            )
            TEST_JOB_HANDLES[id] = handle

    return TEST_JOB_HANDLES[test_job_ID]


# -----------------------------------------------------------------------------


the_job_description = """
    universe = vanilla
    shell = cat common-file input-file-$(ProcID) > output-file-$(ProcID); sleep 10
    MY.CommonInputFiles = "common-file"
    transfer_input_files = input-file-$(ProcID)
    transfer_output_files = output-file-$(ProcID)
    should_transfer_files = YES
    requirements = HasCommonFilesTransfer
    request_cpus = 1
    request_memory = 1024
    log = $(ClusterID).log
    hold = true
"""


@jobs(params={
    "undefined":    f"{the_job_description}",
    "false":        f"{the_job_description}\nMY.RequireCommonFilesTransfer = False\n",
})
def expect_missing_job(request, default_condor):
    return make_test_job(default_condor, request.param)


@jobs(params={
    "true":        f"{the_job_description}\nMY.RequireCommonFilesTransfer = True\n",
})
def expect_present_job(request, default_condor):
    return make_test_job(default_condor, request.param)


class TestSubmitCommand:

    def test_missing(self, expect_missing_job):
        requirements = str(expect_missing_job.clusterad()['Requirements'])
        assert not 'TARGET.HasCommonFilesTransfer >= 2' in requirements


    def test_present(self, expect_present_job):
        requirements = str(expect_present_job.clusterad()['Requirements'])
        assert 'TARGET.HasCommonFilesTransfer >= 2' in requirements



#!/usr/bin/env pytest

# test_dag_halt.py
#
# This tests DAGMan halt DAGMan stops appropriately when
# halted (or resumed). This also verifies that the halt/resume
# interface through the python api DAGMan class and the
# htcondor cli dag noun's halt/resume verbs function correctly.
#
# Note: Most of the complexity is between the test waiting for
#       specific states and the logic in Node A's script to do
#       an action (resume DAG, nothing, hold/release DAG).

from ornithology import *
import htcondor2
import os
import json
import enum
from pathlib import Path
from shutil import rmtree
from time import time as now

#------------------------------------------------------------------
# Enumeration of actions the Node A job can execute
class Actions(str, enum.Enum):
    def __str__(self) -> str:
        return self.value

    NOTHING = "nothing"
    HOLD_N_RELEASE = "hold_and_release"
    RESUME_PY = "resume_py"
    RESUME_TOOL = "resume_tool"

# Timeout for all sections that wait on something
TIMEOUT = 90

# Various files for debugging and timing
CONTINUE_FILE = "dag.continue"
POLL_FILE = "job.running"
ERROR_FILE = "job.err"

# Test specific file that modifies the DAG appropriately
DAG_TEST_DETAILS = "test.details"

# Base DAG file to work with
DAG_FILENAME = "test.dag"
DAG_FILE = """
SUBMIT-DESCRIPTION touch @=desc
    executable = {touch_script}
    arguments  = "$(dag_node_name) job"
    universe   = local
    queue
@desc

JOB A @=NodeA
    executable = {act_on_dag_script}
    arguments  = "$(dag_node_name) $(ACTION) $(DAGManJobId)"
    universe   = local
    getenv     = CONDOR*, PYTHON*, PATH
    error      = {error_file}

    ACTION = {default_action}

    queue
@NodeA

JOB B touch

PARENT A CHILD B

SCRIPT PRE A {touch_script} $NODE pre
SCRIPT POST A {touch_script} $NODE post

SCRIPT PRE B {touch_script} $NODE pre
SCRIPT POST B {touch_script} $NODE post

INCLUDE {test_details}
"""

#==================================================================
# Send halt via python api
def python_halt(condor, ID: int) -> bool:
    success = False
    with condor.use_config():
        dm = htcondor2.DAGMan(ID)
        success, msg = dm.halt("Testing python api DAGMan.halt()")
        print(f"Halt {ID} via python api: {msg}")
    return success

#------------------------------------------------------------------
# Send a pause command via python api
def python_pause(condor, ID: int) -> bool:
    success = False
    with condor.use_config():
        dm = htcondor2.DAGMan(ID)
        success, msg = dm.halt("Testing python api DAGMan.halt() pauses DAG", True)
        print(f"Paused {ID} via python api: {msg}")
        if "paused" not in msg.lower():
            success = False
    return success

#------------------------------------------------------------------
# Send halt command via htcondor noun/verb cli
def tool_halt(condor, id: int) -> bool:
    p = condor.run_command(["htcondor", "dag", "halt", "--reason", "Testing htcondor dag halt", f"{id}"])
    output = p.stdout + p.stderr
    for line in output.split("\n"):
        if f"Halted DAG {id}" in line:
            return True
    return False

#------------------------------------------------------------------
# Class to hold specific test details:
#    - func: Python function to halt send halt command
#    - code: Expected exit code
#    - status: Expected final DAG status
#    - files: List of files (produced by node parts) that we expect to exist
class TestDetails:
    __test__ = False
    def __init__(self, func, code, status, files):
        self.func = func
        assert isinstance(code, int)
        self.code = code
        assert isinstance(status, int)
        self.status = status
        assert isinstance(files, list)
        self.files = files

#==================================================================
# Test cases to execute:
#     NAME : {
#       "inlcude": Test specific DAG file commands to add
#       "details": Test details for execution and verification
#     }
TEST_CASES = {
    "PY_HALT": {
        "details": TestDetails(python_halt, 1, 6, ["A.pre", "A.job", "A.post"]),
    },
    "PY_RESUME": {
        "include": f'VARS A APPEND ACTION="{Actions.RESUME_PY}"',
        "details": TestDetails(python_halt, 0, 0, ["A.pre", "A.job", "A.post", "B.pre", "B.job", "B.post"]),
    },
    "TOOL_HALT": {
        "details": TestDetails(tool_halt, 1, 6, ["A.pre", "A.job", "A.post"]),
    },
    "TOOL_RESUME": {
        "include": f'VARS A APPEND ACTION="{Actions.RESUME_TOOL}"',
        "details": TestDetails(tool_halt, 0, 0, ["A.pre", "A.job", "A.post", "B.pre", "B.job", "B.post"]),
    },
    # Special: Make sure DAGMan runs Final node once halted
    "HALT_RUN_FINAL": {
        "include":
"""
# FINAL Node should run when DAG is halted
# and finished with current work
FINAL F touch

SCRIPT PRE F {touch_script} $NODE pre
SCRIPT POST F {touch_script} $NODE post
""",
        "details": TestDetails(tool_halt, 0, 6, ["A.pre", "A.job", "A.post", "F.pre", "F.job", "F.post"])
    },
    # Special: Verify halt status last beyond crash (or DAGMan being held)
    "HALT_SURVIVE_RECOVERY": {
        "include": f'VARS A APPEND ACTION="{Actions.HOLD_N_RELEASE}"',
        "details": TestDetails(python_halt, 1, 6, ["A.pre", "A.job", "A.post"]),
    },
    # Special: Verify DAGMans ability to pause all progress
    "PAUSE_DAG": {
        "include": f'VARS A APPEND ACTION="{Actions.RESUME_PY}"',
        "details": TestDetails(python_pause, 0, 0, ["A.pre", "A.job", "A.post", "B.pre", "B.job", "B.post"])
    }
}

#==================================================================
# Script to touch a file: touch.py filename file_extension
@action
def write_touch_script(test_dir):
    script = test_dir / "touch.py"
    with open(script, "w") as f:
        f.write(r"""#!/usr/bin/env python3
import sys
from pathlib import Path

if len(sys.argv) != 3:
    sys.exit(1)

Path(f"{sys.argv[1]}.{sys.argv[2]}").touch()

sys.exit(0)
""")
    os.chmod(script, 0o755)
    return script

#------------------------------------------------------------------
# Complex script to do some action(s) to parent DAGMan job once
# the test has informed us that the DAG is halted
# act_on_dag.py node_name action dag_id
@action
def write_action_script(test_dir):
    script = test_dir / "act_on_dag.py"
    with open(script, "w") as f:
        f.write(f"""#!/usr/bin/env python3
import sys
import htcondor2
import os
import subprocess
from pathlib import Path
from time import time as now

def error(msg: str):
    sys.stderr.write(f"{{msg}}\\n")

if len(sys.argv) != 4:
    error(f"Invalid number of args ({{len(sys.argv)}}): Expected 4")
    sys.exit(1)

try:
    NODE = sys.argv[1]
    ACTION = sys.argv[2].lower()
    DAG_ID = int(sys.argv[3])
except Exception as e:
    error(f"Failed to process args: {{e}}")
    sys.exit(2)

# Inform test that node A job is executing
Path("{POLL_FILE}").touch()

# Wait for test to inform that DAG has been sent halt command
start = now()
while True:
    # Just Poll for test to tell us we are ready since test will
    # remove us if/when it times out
    #if now() - start >= ({TIMEOUT} * 2):
    #    error("Timed out waiting on test infrastructure signal file '{CONTINUE_FILE}'")
    #    sys.exit(3)
    if os.path.exists("{CONTINUE_FILE}"):
        break

# Do specified action
if ACTION == "{Actions.HOLD_N_RELEASE}":
    # Hold parent DAG job
    schedd = htcondor2.Schedd()
    result = schedd.act(htcondor2.JobAction.Hold, f"{{DAG_ID}}.0")

    if result["TotalSuccess"] != 1:
        error(f"Failed to put DAG {{DAG_ID}} on hold")
        sys.exit(10)

    # Wait for DAGMan jobs hold event
    with htcondor2.JobEventLog("{DAG_FILENAME}.dagman.log") as JEL:
        start = now()
        held = False
        while True:
            if now() - start >= {TIMEOUT}:
                error("Timed out waiting for hold event after {TIMEOUT} seconds.")
                sys.exit(11)
            for event in JEL.events(stop_after=1):
                assert event.cluster == DAG_ID
                if event.type == htcondor2.JobEventType.JOB_HELD:
                    held = True
                    break
            if held:
                break

    # Release the DAG
    result2 = schedd.act(htcondor2.JobAction.Release, f"{{DAG_ID}}.0")
    if result2["TotalSuccess"] != 1:
        error(f"Failed to release DAG {{DAG_ID}}")
        sys.exit(12)

elif ACTION == "{Actions.RESUME_PY}":
    # Send resume command to DAG via python api
    dm = htcondor2.DAGMan(DAG_ID)
    success, msg = dm.resume()
    if not success:
        error(f"Failed to send resume command to DAG {{DAG_ID}}: {{msg}}")
        sys.exit(20)

elif ACTION == "{Actions.RESUME_TOOL}":
    # Send resume command to DAG via htcondor noun/verb cli
    p = subprocess.run(
        ["htcondor", "dag", "resume", f"{{DAG_ID}}"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        timeout=30,
    )

    success = False
    output = p.stdout.decode() + p.stderr.decode()
    for line in output.split("\\n"):
        if f"Resumed DAG {{DAG_ID}}" in line:
            success = True
            break

    if not success:
        error(f"Failed to resume DAG {{DAG_ID}} from htc-cli:\\n\\n{{output}}")
        sys.exit(30)

elif ACTION == "{Actions.NOTHING}":
    # Do nothing
    pass

else:
    # Unknown action
    error(f"Unknown action provided: {{ACTION}}")
    sys.exit(4)

# Touch our job file
Path(NODE + ".job").touch()

sys.exit(0)
""")
    os.chmod(script, 0o755)
    return script

#------------------------------------------------------------------
@action
def run_dags(default_condor, test_dir, write_touch_script, write_action_script):
    TESTS = dict()

    DAG_MODIFIERS = {
        "touch_script": write_touch_script,
        "act_on_dag_script": write_action_script,
        "test_details": DAG_TEST_DETAILS,
        "error_file": ERROR_FILE,
        "default_action": Actions.NOTHING,
    }

    for test, info in TEST_CASES.items():
        # Create and change into sub-directory named after test case
        os.chdir(str(test_dir))
        exe_dir = test_dir / test
        if exe_dir.exists():
            rmtree(exe_dir)
        os.mkdir(str(exe_dir))
        os.chdir(str(exe_dir))

        # Write base DAG file
        with open(DAG_FILENAME, "w") as f:
            f.write(DAG_FILE.format(**DAG_MODIFIERS))

        # Write test specific DAG include file (or empty dummy)
        with open(DAG_TEST_DETAILS, "w") as f:
            f.write(
                info.get(
                    "include",
                    "# Test is using default information\n"
                ).format(**DAG_MODIFIERS)
            )

        # Submit DAG
        DAG = htcondor2.Submit.from_dag(DAG_FILENAME)
        HANDLE = default_condor.submit(DAG)
        TESTS[test] = (HANDLE, info["details"])

    # Change back to root test dir
    os.chdir(str(test_dir))

    yield TESTS

#------------------------------------------------------------------
@action(params={name: name for name in TEST_CASES})
def test_info(request, run_dags, test_dir):
    return (request.param, run_dags[request.param][0], run_dags[request.param][1])

@action
def test_name(test_info):
    return test_info[0]

@action
def test_handle(test_info):
    return test_info[1]

@action
def test_details(test_info):
    return test_info[2]

#==================================================================
# Magical class to cd into test specific dir and return to root test dir
# via pythons 'with' keyword
class ChangeDir:
    def __init__(self, switch):
        self.orig = None
        self.switch = switch
    def __enter__(self):
        self.orig = os.getcwd()
        os.chdir(self.switch)
    def __exit__(self, exec_type, exec_value, exec_traceback):
        os.chdir(self.orig)
        self.orig = None

#------------------------------------------------------------------
class TestDAGManHalt:
    def test_check_run_actions_and_wait(self, test_name, test_handle, test_details, default_condor):
        with ChangeDir(test_name) as dag_dir:
            # Wait for node A job to inform us that is executing
            start = now()
            while True:
                assert now() - start < TIMEOUT
                if os.path.exists(POLL_FILE):
                    break
            # Run test specific halt command
            assert test_details.func(default_condor, test_handle.clusterid)
            # Inform node A job that we have sent halt command
            Path(CONTINUE_FILE).touch()
            # Wait for DAG job to complete
            assert test_handle.wait(condition=ClusterState.all_complete, timeout=TIMEOUT)


    def test_check_final_status(self, test_name, test_details):
        with ChangeDir(test_name) as dag_dir:
            # Check expected exit code and DAG status to recorded metrics
            with open(f"{DAG_FILENAME}.metrics", "r") as f:
                metrics = json.load(f)
                assert metrics["exitcode"] == test_details.code
                assert metrics["DagStatus"] == test_details.status


    def test_check_expected_files(self, test_name, test_details):
        # Special files we expect to exist but are not part of test verification
        IGNORE = [
            CONTINUE_FILE,
            POLL_FILE,
            ERROR_FILE,
            DAG_TEST_DETAILS,
        ]
        expected = test_details.files
        success = True

        with ChangeDir(test_name) as dag_dir:
            # Iterate over test directory
            for path in os.listdir():
                if path in IGNORE or path.startswith(DAG_FILENAME):
                    # Skip DAG files and explict ignore files
                    continue
                elif path in expected:
                    expected.remove(path)
                else:
                    print(f"Unexpected file discovered: {path}")
                    success = False

            # Verify we found all expected files
            if len(expected) > 0:
                success = False
                for f in expected:
                    print(f"Missing expected file: {f}")

        assert success



#!/usr/bin/env pytest

import os
import time

from ornithology import (
    action,
    write_file,
    format_script,
    ClusterState,
)

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

@action
def path_to_the_job_script(test_dir):
    script = """
    #!/usr/bin/python3

    import sys

    print("standard output", file=sys.stdout)
    print("standard error", file=sys.stderr)
    sys.exit(0)
    """

    path = test_dir / "logging.py"
    write_file(path, format_script(script))

    return path


@action
def the_job_description(path_to_the_job_script):
    return {
        # Setting this means that all terminal job states leave the job
        # in the queue until it's removed, which is convenient.
        "LeaveJobInQueue":              "true",

        "executable":                   path_to_the_job_script.as_posix(),
        "transfer_executable":          "true",
        "should_transfer_files":        "true",
        "when_to_transfer_output":      "ON_EXIT",
        "success_exit_code":            1,
        "max_retries":                  0,
    }


@action
def the_job_handle(default_condor, test_dir, the_job_description):
    job_description = {
        ** the_job_description,
        ** {
            "log":              test_dir / "logged_job.log",
            "output":           test_dir / "the_output_log",
            "error":            test_dir / "the_error_log",
        }
    }

    job_handle = default_condor.submit(
        description = job_description,
        count = 1,
    )

    yield job_handle

    job_handle.remove()


@action
def the_completed_job_handle(the_job_handle):
    the_job_handle.wait(
        timeout = 60,
        condition = ClusterState.all_complete,
        fail_condition = ClusterState.any_held,
    )

    return the_job_handle


class TestSuccessExitCode:

    def test_transfer(self, test_dir, the_completed_job_handle):
        the_output_log = test_dir / "the_output_log"
        assert the_output_log.is_file()
        assert the_output_log.read_text() == "standard output\n"

        the_error_log = test_dir / "the_error_log"
        assert the_error_log.is_file()
        assert the_error_log.read_text() == "standard error\n"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------

project = "Ornithology"
copyright = "1990-2020, Center for High Throughput Computing, Computer \
Sciences Department, University of Wisconsin-Madison, Madison, WI, US. \
Licensed under the Apache License, Version 2.0."

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.graphviz",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
# exclude_patterns = ["**/.ipynb_checkpoints"]

pygments_style = "colorful"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]

# -- Extension Configuration -------------------------------------------------

autoclass_content = "both"

intersphinx_mapping = {
    'python': ("https://docs.python.org/3/", None),
}



from typing import (
    Union,
    List,
    Optional,
)

from ._common_imports import (
    classad,
    Collector,
    DaemonType,
)

from ._drain_type import DrainType
from ._vacate_type import VacateType
from ._completion_type import CompletionType
from ._history_src import HistorySrc

from .htcondor2_impl import (
    _startd_drain_jobs,
    _startd_vacate_slots,
    _startd_cancel_drain_jobs,
    _history_query,
)


class Startd():
    """
    The startd client.  Used to access the startd history and to control
    `draining <https://htcondor.readthedocs.io/en/latest/admin-manual/cm-configuration.html#defragmenting-dynamic-slots>`_.
    """

    def __init__(self, location : Optional[classad.ClassAd] = None):
        """
        :param location:  A ClassAd with a ``MyAddress`` attribute, such as
            might be returned by :meth:`htcondor2.Collector.locate`.  :py:obj:`None` means the
            the local daemon.
        """
        if location is None:
            c = Collector()
            location = c.locate(DaemonType.Startd)

        if not isinstance(location, classad.ClassAd):
            raise TypeError("location must be a ClassAd")

        self._addr = location['MyAddress']
        # We never actually use this for anything.
        # self._version = location['CondorVersion']


    # In version 1, check_expr and start_expr could also be `ExprTree`s.
    def drainJobs(self,
      drain_type : DrainType = DrainType.Graceful,
      on_completion : CompletionType = CompletionType.Nothing,
      check_expr : Optional[str] = None,
      start_expr : Optional[str] = None,
      reason : Optional[str] = None,
    ) -> str:
        """
        Request that this startd begin draining.

        :param drain_type:  How much time, if any, jobs are given to finish.
        :param on_completion:  What to do when the startd finishes draining.
        :param check_expr:  An expression which must evaluate to :py:obj:`True`
            in the context of each slot ad for draining to begin.  If
            :py:obj:`None`, treated as ``True``.
        :param start_expr:  The value of the
            `START <https://htcondor.readthedocs.io/en/latest/admin-manual/configuration-macros.html#START>`_
            expression while the startd is draining.  If
            :py:obj:`None`, use the configured default
            (`DEFAULT_DRAINING_START_EXPR <https://htcondor.readthedocs.io/en/latest/admin-manual/configuration-macros.html#DEFAULT_DRAINING_START_EXPR>`_).
        :param reason:  A string describing the reason for draining.

        :return:  An opaque request ID only useful for cancelling this drain
            via :meth:`cancelDrainJobs`.
        """
        if check_expr is not None and not isinstance(check_expr, str):
            raise TypeError("check_expr must be a string")
        if start_expr is not None and not isinstance(start_expr, str):
            raise TypeError("start_expr must be a string")
        return _startd_drain_jobs(self._addr,
          int(drain_type), int(on_completion), check_expr, start_expr, reason
        )

    def vacateSlot(self, slot_name : Optional[str] = None, vacate_fast : VacateType = VacateType.Graceful) -> None:
        """
        Evict the job running on an active slot. Eviction is graceful unless
        vacate_fast is set to VacateType.Fast.

        :param slot_name:  The name of the slot.
        :param vacate_fast:  :py:obj:`Fast` to evict fast. :py:obj:`None` or :py:obj:`Graceful` to evict graceful.
        """
        if slot_name is None or slot_name == "":
            raise ValueError("slot_name must be a non-empty string");
        _startd_vacate_slots(self._addr, slot_name, vacate_fast)

    def vacate(self, vacate_fast : VacateType = VacateType.Graceful) -> None:
        """
        Evict all jobs running on active slots. Eviction is graceful unless
        vacate_fast is set to VacateType.Fast.

        :param vacate_fast:  :py:obj:`Fast` to evict fast. :py:obj:`None` or :py:obj:`Graceful` to evict graceful.
        """
        _startd_vacate_slots(self._addr, None, vacate_fast)


    def cancelDrainJobs(self, request_id : Optional[str] = None) -> None:
        """
        Cancel a draining request.

        :param request_id:  An opaque drain request ID returned by a
            previous call to :meth:`drainJobs`.  If :py:obj:`None`,
            cancel all draining requests for this startd.
        """
        _startd_cancel_drain_jobs(self._addr, request_id)


    # Totally undocumented in version 1.
    def history(self,
        constraint : Optional[Union[str, classad.ExprTree]] = None,
        projection : List[str] = [],
        match : int = -1,
        since : Union[int, str, classad.ExprTree] = None,
    ) -> List[classad.ClassAd]:
        """
        Query this startd's job history.

        :param constraint:  A query constraint.  Only jobs matching this
            constraint will be returned.  :py:obj:`None` will match all jobs.
        :param projection:  A list of job attributes.  These attributes will
            be returned for each job in the list.  (Others may be as well.)
            The default (an empty list) returns all attributes.
        :param match:  The maximum number of ads to return.  The default
            (``-1``) is to return all ads.
        :param since:  A cluster ID, job ID, or expression.  If a cluster ID
            (passed as an integer) or job ID (passed as a string in the format
            ``{clusterID}.{procID}``), only jobs recorded in the history
            file after (and not including) the matching ID will be
            returned.  If an expression (including strings that aren't
            job IDs), jobs will be returned, most-recently-recorded first,
            until the expression becomes true; the job making the expression
            become true will not be returned.  Thus, ``1038`` and
            ``"clusterID == 1038"`` return the same set of jobs.

            If :py:obj:`None`, return all (matching) jobs.
        :return:  A list of `job ClassAds <https://htcondor.readthedocs.io/en/latest/classad-attributes/job-classad-attributes.html>`_.
        """
        projection_string = ",".join(projection)

        if isinstance(since, int):
            since = f"ClusterID == {since}"
        elif isinstance(since, str):
            pattern = re.compile(r'(\d+).(\d+)')
            matches = pattern.match(since)
            if matches is None:
                raise ValueError("since string must be in the form {clusterID}.{procID}")
            since = f"ClusterID == {matches[0]} && ProcID == {matches[1]}"
        elif isinstance(since, classad.ExprTree):
            since = str(since)
        elif since is None:
            since = ""
        else:
            raise TypeError("since must be an int, string, or ExprTree")

        if constraint is None:
            constraint = ""

        return _history_query(self._addr,
            str(constraint), projection_string, int(match), since,
            # Ad Type Filter
            "",
            HistorySrc.StartdJob,
            # GET_HISTORY
            429,
        )




# Copyright 2024 HTCondor Team, Computer Sciences Department,
# University of Wisconsin-Madison, WI.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import json
import logging
import traceback
import re

from itertools import chain

from adstash.ad_sources.generic import GenericAdSource

import htcondor2 as htcondor
import classad2 as classad


class ScheddTransferEpochHistorySource(GenericAdSource):


    def __init__(self, *args, **kwargs):
        self.text_attrs = {
            "AttemptError",
            "ErrorString",
            "TransferError",
        }
        self.indexed_keyword_attrs = {
            "Endpoint",
            "ErrorType",
            "ErrorMessage",
            "FailedName",
            "FailureType",
            "DebugErrorType",
            "GLIDEIN_ResourceName",
            "GLIDEIN_Site",
            "GlobalJobId",
            "HttpCacheHitOrMiss",
            "HttpCacheHost",
            "Owner",
            "PelicanClientVersion",
            "ProjectName",
            "ScheddName",
            "ServerVersion",
            "StartdName",
            "StartdSlot",
            "TransferHostName",
            "TransferLocalMachineName",
            "TransferProtocol",
            "TransferType",
        }
        self.noindex_keyword_attrs = {
            "TransferFileName",
            "TransferUrl",
        }
        self.float_attrs = {
            "AttemptTime",
            "ConnectionTimeSeconds",
            "TimeToFirstByte",
        }
        self.int_attrs = {
            "Attempt",
            "Attempts",
            "AttemptFileBytes",
            "ClusterId",
            "ErrorCode",
            "LibcurlReturnCode",
            "NumShadowStarts",
            "PelicanErrorCode",
            "ProcId",
            "TransferFileBytes",
            "TransferHttpStatusCode",
            "TransferTotalBytes",
            "TransferTries",
        }
        self.date_attrs = {
            "AttemptEndTime",
            "EpochWriteDate",
            "RecordTime",
            "TransferEndTime",
            "TransferStartTime",
        }
        self.bool_attrs = {
            "FinalAttempt",
            "Retryable",
            "TransferSuccess",
        }
        self.nested_attrs = set()
        self.known_attrs = (
            self.text_attrs |
            self.indexed_keyword_attrs |
            self.noindex_keyword_attrs |
            self.float_attrs |
            self.int_attrs |
            self.date_attrs |
            self.bool_attrs |
            self.nested_attrs )
        self.known_attr_map = {attr.lower(): attr for attr in self.known_attrs}
        self.dynamic_templates = [
            {
                "strings_as_keywords": {  # Store unknown strings as keywords
                    "match_mapping_type": "string",
                    "mapping": {"type": "keyword", "index": "false", "norms": "false", "ignore_above": 256},
                }
            },
        ]
        super().__init__(*args, **kwargs)


    def fetch_ads(self, schedd_ad, max_ads=10000, projection=set()):
        if projection:
            logging.warning(f"Custom projections are ignored for transfer history ads")

        history_kwargs = {}
        if max_ads > 0:
            history_kwargs["match"] = max_ads

        ckpt = self.checkpoint.get(f"Transfer Epoch {schedd_ad['Name']}")
        if ckpt is None:
            logging.warning(f"No transfer epoch checkpoint found for schedd {schedd_ad['Name']}, getting all ads available.")
        else:
            since_exprs = []
            for attr in ("EpochWriteDate", "ClusterId", "ProcId", "NumShadowStarts",):
                if attr in ckpt:
                    since_exprs.append(f"({attr} == {ckpt[attr]})")
            since_expr = " && ".join(since_exprs)
            history_kwargs["since"] = classad.ExprTree(since_expr)
            logging.warning(f"Getting transfer epoch ads from {schedd_ad['Name']} since {since_expr}.")
        schedd = htcondor.Schedd(schedd_ad)
        return schedd.jobEpochHistory(constraint=True, projection=[], ad_type="TRANSFER", **history_kwargs)


    def unique_doc_id(self, doc):
        """
        To uniquely identify documents (not jobs)
        """
        schedd_name = doc.get("ScheddName", "unknown") or "unknown"
        job_id = f"{doc.get('ClusterId') or 0}.{doc.get('ProcId') or 0}"
        shadow_starts = doc.get("NumShadowStarts", 0)
        xfer_protocol = doc.get("TransferProtocol", "unknown")
        xfer_type = doc.get("TransferType", "unknown")
        date = doc.get("RecordTime", 0)
        attempt = f"{doc.get('Attempt', 0)}_{doc.get('Attempts', 1)}"
        return f"{schedd_name}#{job_id}#{shadow_starts}#{xfer_protocol}#{xfer_type}#{date}#{attempt}"


    def normalize(self, attr, value):
        attr = self.known_attr_map.get(attr.lower(), attr.lower())
        if value is None:
            pass
        else:
            try:
                if attr not in self.known_attrs:
                    value = str(value)
                elif attr in (self.int_attrs | self.date_attrs):
                    try:
                        value = int(value)
                    except TypeError:
                        value = str(value)
                        attr = f"{attr.lower()}_string"
                elif attr in self.float_attrs:
                    try:
                        value = float(value)
                    except TypeError:
                        value = str(value)
                        attr = f"{attr.lower()}_string"
                elif attr in self.bool_attrs:
                    try:
                        value = bool(value)
                    except TypeError:
                        value = str(value)
                        attr = f"{attr.lower()}_string"
                else:
                    value = str(value)
            except Exception as e:
                attr = f"{attr}_error"
                value = f"{e.__class__.__name__}: {str(e)}"
            if isinstance(value, str) and attr not in self.text_attrs and len(value) > 255:
                value = f"{value[:253]}..."
        return attr, value


    def expand_plugin_result_ads(self, ads, my_attr_name=None):
        debug_results = []
        error_results = iter(())
        result = {}
        for ad in ads:
            for attr, value in ad.items():
                if attr.lower() == "developerdata":
                    debug_results = self.expand_debug_ad(value)
                elif attr.lower() == "transfererrordata":
                    error_results = self.expand_error_ads(value)
                else:
                    attr, value = self.normalize(attr, value)
                    result[attr] = value
            if result.get("TransferProtocol") is None:
                result["TransferProtocol"] = result.get("TransferUrl", "unknown:").split("+", maxsplit=1)[-1].split(":", maxsplit=1)[0]
            if result.get("TransferType") is None:
                result["TransferType"] = {"InputPluginResultList": "download", "OutputPluginResultList": "upload"}.get(my_attr_name, "unknown")
            if not debug_results:
                result["Attempts"] = 1
                result["Attempt"] = 0
                result["FinalAttempt"] = True
                try:  # add error data if it exists
                    result.update(next(error_results))
                except StopIteration:
                    pass
                yield result
            else:
                for debug_result in debug_results:
                    try:
                        error_result = next(error_results)
                    except StopIteration:
                        error_result = {}
                    if debug_result.get("FinalAttempt"):  # merge and return entire ad on final attempt
                        try:
                            yield debug_result | result | error_result
                        except TypeError:  # backwards compat
                            yield {**debug_result, **result}
                    else:  # otherwise only add identifying attrs
                        for attr in ("TransferProtocol", "TransferType", "TransferUrl"):
                            debug_result[attr] = result.get(attr)
                        yield debug_result | error_result


    def expand_debug_ad(self, ad):
        result = {}
        try:
            attempts = int(ad.get("Attempts"))
        except TypeError:
            attempts = None

        if attempts is None:  # Attempts was unspecified in source ad
            result["Attempts"] = 1
            result["Attempt"] = 0
            result["FinalAttempt"] = True
            for attr, value in ad.items():
                attr, value = self.normalize(attr, value)
                result[attr] = value
            yield result
            return

        if attempts == 0:  # Attempts was actually 0 in source ad
            result["Attempts"] = 0
            result["Attempt"] = 0
            result["FinalAttempt"] = True
            for attr, value in ad.items():
                attr, value = self.normalize(attr, value)
                result[attr] = value
            yield result
            return

        attempt_attr_re = re.compile(r"(\D+)(\d+)$")
        attempt_attrs = set()
        for attr, value in ad.items():
            attempt_attr_match = attempt_attr_re.match(attr)
            if attempt_attr_match:
                attempt_attrs.add(attempt_attr_match.group(1))
            else:
                attr, value = self.normalize(attr, value)
                result[attr] = value

        for attempt in range(attempts):
            attempt_result = {
                "Attempt": attempt,
                "FinalAttempt": attempt + 1 == attempts,
            }
            for attr in attempt_attrs:
                attempt_attr = f"{attr}{attempt}"
                value = ad.get(attempt_attr)
                if value is None:
                    continue
                if attr.lower().startswith("transfer"):
                    attr = f"Attempt{attr[len('transfer'):]}"
                attr, value = self.normalize(attr, value)
                attempt_result[attr] = value
            try:
                yield result | attempt_result
            except TypeError:  # backwards compat
                yield {**result, **attempt_result}


    def expand_error_ads(self, ads):
        for ad in ads:
            error_result = {}
            for attr, value in ad.items():
                if attr.lower() == "developerdata":
                    for debug_attr, debug_value in value.items():
                        debug_attr, debug_value = self.normalize(debug_attr, debug_value)
                        if debug_attr in ad:
                            debug_attr = f"Debug{debug_attr}"
                        error_result[debug_attr] = debug_value
                else:
                    attr, value = self.normalize(attr, value)
                    error_result[attr] = value
            yield error_result


    def to_json_list(self, ad, return_dict=False):
        result = {}

        # Technically, there could be multiple "*PluginResultList" attrs in an epoch ad.
        # plugin_results is a list of generators of dicts, which will be consumed using itertools.chain().
        plugin_results = []

        # Python None converts to JSON null. Helpfully, Elasticsearch will ignore null values.
        result["RecordTime"] = ad.get("EpochWriteDate", int(time.time()))
        result["ScheddName"] = ad.get("GlobalJobId", "").split("#", maxsplit=1)[0] or None
        result["ClusterId"] = ad.get("ClusterId", ad.get("GlobalJobId", "#.").split("#")[1].split(".")[0]) or None
        result["ProcId"] = ad.get("ProcId", ad.get("GlobalJobId", "#.").split("#")[1].split(".")[-1]) or None
        result["StartdSlot"] = ad.get("RemoteHost", "").split("@", maxsplit=1)[0] or None
        result["StartdName"] = ad.get("RemoteHost", "").split("@", maxsplit=1)[-1] or None
        for attr, value in result.items():
            attr, value = self.normalize(attr, value)
            result[attr] = value
        for attr, value in ad.items():
            if attr.endswith("PluginResultList"):
                plugin_results.append(self.expand_plugin_result_ads(value, my_attr_name=attr))
            else:
                attr, value = self.normalize(attr, value)
                result[attr] = value
        for plugin_result in chain(*plugin_results or [{}]):
            if return_dict:
                try:
                    yield result | plugin_result
                except TypeError:  # backwards compat
                    yield {**result, **plugin_result}
            else:
                try:
                    yield json.dumps(result | plugin_result)
                except TypeError:  # backwards compat
                    yield json.dumps({**result, **plugin_result})


    def process_ads(self, interface, ads, schedd_ad, metadata={}, chunk_size=0, **kwargs):
        starttime = time.time()
        chunk = []
        schedd_checkpoint = None
        ads_posted = 0
        for ad in ads:
            try:
                dict_ads = self.to_json_list(ad, return_dict=True)
            except Exception as e:
                message = f"Failure when converting document in {schedd_ad['name']} transfer epoch history: {str(e)}"
                exc = traceback.format_exc()
                message += f"\n{exc}"
                logging.warning(message)
                continue

            # Unfortunately, the schedd history is in reverse chronological order,
            # therefore the checkpoint should be set to the first ad that is returned.
            # Here, we assume that the interface is responsible for de-duping ads
            # and only update the checkpoint after the full history queue is pushed
            # through by returning the new checkpoint at the end.

            if schedd_checkpoint is None:  # set checkpoint based on first parseable ad
                schedd_checkpoint = {}
                for attr in ("EpochWriteDate", "ClusterId", "ProcId", "NumShadowStarts",):
                    if attr in ad:
                        schedd_checkpoint[attr] = ad[attr]
            for dict_ad in dict_ads:
                dict_ad["ScheddName"] = schedd_ad["name"]
                chunk.append((self.unique_doc_id(dict_ad), dict_ad,))

            if (chunk_size > 0) and (len(chunk) >= chunk_size):
                logging.debug(f"Posting {len(chunk)} transfer epoch ads from {schedd_ad['Name']}.")
                result = interface.post_ads(chunk, metadata=metadata, ad_source=self, **kwargs)
                ads_posted += result["success"]
                yield None  # don't update checkpoint yet, per note above
                chunk = []

        if len(chunk) > 0:
            logging.debug(f"Posting {len(chunk)} transfer epoch ads from {schedd_ad['Name']}.")
            result = interface.post_ads(chunk, metadata=metadata, ad_source=self, **kwargs)
            ads_posted += result["success"]

        endtime = time.time()
        logging.warning(f"Schedd {schedd_ad['Name']} transfer epoch history: response count: {ads_posted}; upload time: {(endtime-starttime)/60:.2f} min")
        yield schedd_checkpoint  # finally update checkpoint



#!/usr/bin/env pytest

import os
import pytest
from pathlib import Path
import subprocess
import htcondor2

#
# Runs the C++ test_starter_guidance.exe, a unit test.
#
class TestStarterGuidance:

    def test_unit_test(self, test_dir):
        the_path = Path(htcondor2.param['LIBEXEC']) / "test_starter_guidance.exe"
        the_env = {
            ** os.environ,
            "_CONDOR_LOG": test_dir.as_posix(),
        }

        rv = subprocess.run(
            [the_path, "-f", "-t"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            timeout=20,
            universal_newlines=True,
            env=the_env,
        )

        print(rv.stdout)
        print(rv.stderr)

        assert rv.returncode == 0



import os
import sys
import re
from pathlib import Path

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.nodes import split_explicit_title, process_index_entry, set_role_source_info
from htc_helpers import *

SPECIAL_CASE_KNOBS = ["<SUBSYS>"]
TEMPLATE_FILE = "introduction-to-configuration"

CONFIG_REGEX = {}
CONFIG_KNOBS = {}
TEMPLATES = {}

def find_conf_knobs(dir: str):
    knobs = {}
    regex_map = {}

    base = Path(dir)
    config_dir = base / "admin-manual" / "configuration"

    for path in config_dir.iterdir():
        if path.suffix != ".rst":
            continue

        url = path.relative_to(base).with_suffix(".html")
        with open(path, "r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                for knob in get_all_defined_role("macro-def", line):
                    if "[" in knob:
                        info = knob.find("[")
                        knob = knob[:info]
                    # Store macro for warning check
                    # Handle special case knobs
                    if knob in SPECIAL_CASE_KNOBS:
                        knobs[knob] = url
                    # Check if knob requires regex matching
                    elif "*" in knob:
                        regex = rf"{knob.replace('*', '(.+)')}"
                        regex_map.update({regex : (knob, url)})
                    elif "<" in knob:
                        temp = knob
                        while "<" in temp:
                            start = temp.find("<")
                            end = temp.find(">")
                            temp = regex = rf"{temp.replace(temp[start:end+1], '(.+)')}"
                        # Don't allow (.+) to be a regex match option
                        if regex != "(.+)":
                            regex_map.update({regex : (knob, url)})
                    # Store whatever is left (if not already included)
                    else:
                        knobs[knob] = url

    sorted_regex_map = {k: regex_map[k] for k in sorted(regex_map.keys(), reverse=True)}

    return (knobs, sorted_regex_map)

def find_templates(dir: str):
    templates = {}
    intro_file = Path(dir) / "admin-manual" / f"{TEMPLATE_FILE}.rst"
    with open(intro_file, "r", encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if "config-template" in line:
                start = line.find("`") + 1
                cat_start = line.find("<")
                cat_end = line.rfind(">")
                if cat_start == -1:
                    end = line.find("`")
                    category = line[start:end]
                    templates.update({category : []})
                else:
                    category = line[cat_start + 1:cat_end]
                    end = line.find("(") if "(" in line else cat_start
                    template = line[start:end].upper()
                    if category not in templates.keys():
                        templates.update({category : [template]})
                    else:
                        templates[category].append(template)
    return templates

def generate_old_redirect_page(dir: str):
    """Write javascript to redirect old monolithic configuration macro URLs to new locations"""
    gen_dir = Path(dir) / "_static" / "generated" / "js"
    try:
        gen_dir.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        warn(f"Failed to ensure/create '{gen_dir}': {e}")

    javascript = gen_dir / "config-macro-redirect.js"
    with open(javascript, "w") as f:
        f.write("""
// Map of tag -> URL end
const configuration = new Map([
""")

        for macro, path in CONFIG_KNOBS.items():
            f.write(f'\t["{macro.upper()}", "{path}"],\n')

        f.write("""]);


const config_regex = new Map([
""")

        for regex, info in CONFIG_REGEX.items():
            f.write(f'\t[/{regex.upper()}/, "{info[1]}"],\n')

        f.write("""]);

// Magic Redirect function
function configRedirect() {
	var end = window.location.href.indexOf("admin-manual");
	var url_root = window.location.href.substring(0, end);
	var redirect_url = url_root + "admin-manual/configuration/index.html";

	if (window.location.hash.length > 0) {
		var macro = decodeURI(window.location.hash.slice(1)).toUpperCase();

		if (configuration.has(macro)) {
			// Specific macro found
			redirect_url = url_root + configuration.get(macro) + "#" + macro;
		} else {
			// Macro matches regex configuration option
			for (const check of config_regex.keys()) {
				if (check.test(macro)) {
					redirect_url = url_root + config_regex.get(check) + "#" + macro;
					break;
				}
			}
		}
	}

	window.location.replace(redirect_url);
}

window.addEventListener('DOMContentLoaded', configRedirect());
""")

def dump(obj):
    for attr in dir(obj):
        print("obj.%s = %r" % (attr, getattr(obj, attr)))

def macro_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    root_dir = get_rel_path_to_root_dir(inliner)[:-1]
    macro_name, macro_index = custom_ext_parser(text)

    # Handle reference to configuration template
    if macro_name.strip().lower()[:4] == "use ":
        info = macro_name[4:].strip()
        if ":" in info:
            category, template = info.split(":")[:2]
            category = category.strip().upper()
            template = template.strip().upper()
            if category not in TEMPLATES.keys():
                warn(f"Config template category '{category}' is not defined or a typo exists.")
            elif template not in TEMPLATES[category]:
                warn(f"Config template '{category}:{template}' is not defined or a typo exists.")
            ref = category + ":" + template
        else:
            ref = info.upper()
            if ref not in TEMPLATES.keys():
                warn(f"Config template category '{ref}' is not defined or a typo exists.")
        ref_link = f"href=\"{root_dir}/admin-manual/{TEMPLATE_FILE}.html#" + str(ref) + "\""
    # Handle reference to normal configuration knob
    else:
        url_path = CONFIG_KNOBS.get(macro_name, "admin-manual/configuration-macros.html")
        specifier, macro = macro_name.split(".", 1) if "." in macro_name else (None, macro_name)
        ref = macro
        if macro not in CONFIG_KNOBS.keys():
            regex_match = False
            for r in CONFIG_REGEX.keys():
                if re.match(r, macro, flags=re.IGNORECASE):
                    regex_match = True
                    ref, url_path = CONFIG_REGEX[r]
                    break
            # If here then not in pure defined list or matched a recorded regex
            if not regex_match:
                docname = inliner.document.settings.env.docname
                warn(f"{docname}:{lineno} | Config knob '{macro}' not found in defined list. Either a typo or knob needs definition.")
        ref_link = f"href=\"{root_dir}/{url_path}#" + str(ref) + "\""
    return make_ref_and_index_nodes(name, macro_name, macro_index,
                                    ref_link, rawtext, inliner, lineno, options)

def setup(app):
    global CONFIG_KNOBS
    global CONFIG_REGEX
    global TEMPLATES

    CONFIG_KNOBS, CONFIG_REGEX = find_conf_knobs(app.srcdir)
    TEMPLATES = find_templates(app.srcdir)

    generate_old_redirect_page(app.srcdir)

    app.add_role("macro", macro_role)




#!/usr/bin/env pytest

# Test that we can create sub-cgroups.  If not, skip the test
# if so, turn on the secret unprivileged cgroup option,
# and submit a job that uses more memory than we have
# requested.  Verify that it goes on hold.

import logging
import pytest
import os

import htcondor2 as htcondor
from ornithology import *

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Standup a condor with user-space cgroup v2 enabled
@standup
def condor(test_dir):
    with Condor(test_dir / "condor", config={
        "CREATE_CGROUP_WITHOUT_ROOT": "true",
        "LOCAL_UNIVERSE_CGROUP_ENFORCEMENT": "true",
        }) as condor:
        yield condor

# Standup a condor with user-space cgroup v2 enabled and job hook
@standup
def condor_with_hook(test_dir):
    with Condor(test_dir / "condor_with_hooks", config={"CREATE_CGROUP_WITHOUT_ROOT": "true",
                                                        "STARTER_DEFAULT_JOB_HOOK_KEYWORD": "TEST",
                                                        "TEST_HOOK_PREPARE_JOB": test_dir / "hook_starter_test_prepare.sh"}) as condor_with_hook:
        yield condor_with_hook

# Standup a condor with user-space cgroup v2 enabled and SUBMIT_REQUEST_MISSING_UNITS=error
@standup
def condor_with_units_error(test_dir):
    with Condor(test_dir / "condor_units_error", config={"CREATE_CGROUP_WITHOUT_ROOT": "true",
                                                          "SUBMIT_REQUEST_MISSING_UNITS": "error"}) as condor_units:
        yield condor_units

# The actual job will allocate 200 Mb in python, but request 50Mb.
# The 50Mb might get rounded up to 128, so be careful.
@action
def job_script_contents():
    return format_script( """
#!/usr/bin/env python3

import os
import time

two_hundred_meg = bytearray(200 * 1024 * 1024)
for i in range(10):
    time.sleep(1)
exit(0)
""")

@action
def test_script(test_dir, job_script_contents):
    test_script = test_dir / "job_script.py"
    write_file(test_script, job_script_contents)
    return test_script

@action
def test_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "oom_log",
            "keep_claim_idle": "300",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def completed_test_job(condor, test_job_hash):
    ctj = condor.submit(
        {**test_job_hash}, count=1
    )
    assert ctj.wait(
        condition=ClusterState.all_held,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_complete,
    )
    return ctj

@action
def escaping_job_hash(test_dir, path_to_python, test_script):
    # A job which spawns "sleep 1234", with all the condor tracking env vars removed
    # and nohup'd, to avoid all non-cgroup tracking mechanisms.  Writes the pid
    # to pid file, which lives in the test_dir.  If the pid in pidfile exists
    # after the job exits, the sleep escaped the job, and the test should fail.
    return {
            "shell": "nohup /usr/bin/env -i /usr/bin/sleep 1234 & echo $! > pidfile",
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "escaped_log",
            "keep_claim_idle": "300",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def escaping_test_job(condor, escaping_job_hash):
    ej = condor.submit(
        {**escaping_job_hash}, count=1
    )
    assert ej.wait(
        condition=ClusterState.all_complete,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return ej

@action
def isolated_job_hash(test_dir, path_to_python):
    # A job which cats the content of its own cgroup.procs.  Because we
    # exec, the shell has gone away, so the cgroup should only
    # contain the cat process, which verifies that the job is isolated into
    # its own cgroup.
    return {
            "shell": "exec /usr/bin/cat /sys/fs/cgroup$(awk -F: '/^0:/ {print $3}' < /proc/self/cgroup)/cgroup.procs > cgroup.procs",
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "isolated_log",
            "keep_claim_idle": "300",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }


@action
def isolated_test_job(condor, isolated_job_hash):
    ij = condor.submit(
        {**isolated_job_hash}, count=1
    )
    assert ij.wait(
        condition=ClusterState.all_complete,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return ij

@action
def slice_scope_job_hash(test_dir, path_to_python):
    # A job which cats the content of /proc/self/cgroup.
    # This can be interrogated to verify that the job is the right cgroup.  
    return {
            "shell": "/usr/bin/cat /proc/self/cgroup > cgroup",
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "slice_scope_log",
            "keep_claim_idle": "300",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def slice_scope_test_job(condor, slice_scope_job_hash):
    ssj = condor.submit(
        {**slice_scope_job_hash}, count=1
    )
    assert ssj.wait(
        condition=ClusterState.all_complete,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return ssj

@action
def memory_detection_job_hash(test_dir, condor, path_to_python):
    # A job which runs condor_config_val DETECTED_MEMORY.
    # This should return the memory limit of the cgroup
    bin_dir = condor.run_command( ['condor_config_val', 'BIN']).stdout
    return {
            "shell": f"{bin_dir}/condor_config_val DETECTED_MEMORY",
            "environment": "CONDOR_CONFIG=/dev/null",
            "universe": "vanilla",
            "output": "detected_memory",
            "error": "error",
            "log": "detected_memory.log",
            "keep_claim_idle": "300",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def hook_script_contents():
    return format_script( """
#!/bin/bash
echo 'HookStatusCode = 0'
exit 0
""")

@action
def hook_script(test_dir, hook_script_contents):
    hook_script = test_dir / "hook_starter_test_prepare.sh"
    write_file(hook_script, hook_script_contents)
    os.chmod(hook_script, 0o755)
    return hook_script

@action
def hook_test_job_hash(test_dir, hook_script, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "hook_log",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def memory_detection_test_job(condor, memory_detection_job_hash):
    mmj = condor.submit(
        {**memory_detection_job_hash}, count=1
    )
    assert mmj.wait(
        condition=ClusterState.all_complete,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return mmj

@action
def hook_test_job(condor_with_hook, hook_test_job_hash):
    htj = condor_with_hook.submit(
        {**hook_test_job_hash}, count=1
    )
    assert htj.wait(
        condition=ClusterState.all_held,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_complete,
    )
    return htj

@action
def retry_memory_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "retry_log",
            "keep_claim_idle": "300",
            "request_memory": "50mb",
            "retry_request_memory": "400mb",
            "request_cpus": "1",
            "request_disk": "200mb"
            }

@action
def retry_memory_job(condor, retry_memory_job_hash):
    ctj = condor.submit(
        {**retry_memory_job_hash}, count=1
    )
    assert ctj.wait(
        condition=ClusterState.all_complete,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_held,
    )
    return ctj

@action
def retry_memory_no_units_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "retry_no_units_log",
            "keep_claim_idle": "300",
            "request_memory": "50mb",
            "retry_request_memory": "400",  # No units - should fail
            "request_cpus": "1",
            "request_disk": "200mb"
            }

@action
def retry_memory_max_no_units_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "retry_max_no_units_log",
            "keep_claim_idle": "300",
            "request_memory": "50mb",
            "retry_request_memory_max": "500",  # No units - should fail
            "request_cpus": "1",
            "request_disk": "200mb"
            }

@action
def retry_memory_with_units_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "retry_with_units_log",
            "keep_claim_idle": "300",
            "request_memory": "50mb",
            "retry_request_memory": "400mb",  # With units - should succeed
            "request_cpus": "1",
            "request_disk": "200mb"
            }

@action
def retry_memory_max_with_units_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "vanilla",
            "output": "output",
            "error": "error",
            "log": "retry_max_with_units_log",
            "keep_claim_idle": "300",
            "request_memory": "50mb",
            "retry_request_memory_max": "500mb",  # With units - should succeed
            "request_cpus": "1",
            "request_disk": "200mb"
            }

@action
def local_universe_job_hash(test_dir, path_to_python, test_script):
    return {
            "executable": path_to_python,
            "arguments": test_script,
            "universe": "local",
            "output": "local_output",
            "error": "local_error",
            "log": "local_oom_log",
            "request_memory": "50m",
            "request_cpus": "1",
            "request_disk": "200m"
            }

@action
def completed_local_test_job(condor, local_universe_job_hash):
    clj = condor.submit(
        {**local_universe_job_hash}, count=1
    )
    assert clj.wait(
        condition=ClusterState.all_held,
        timeout=120,
        verbose=True,
        fail_condition=ClusterState.any_complete,
    )
    return clj

# These tests only works if we start in a writeable (delegated)
# cgroup.  Return true if this is so
def CgroupIsWriteable():
    if os.environ.get("HTCONDOR_TEST_USE_SUDO") == "1":
        # When using sudo, we should have root
        return True

    with open('/proc/self/cgroup') as fd:
        proc_self_cgroup = fd.read()

    # proc_self_cgroup on a cgroup v2 system should contain
    # ::/path/to/cgroup/without/leading/sys/fs/cgroup
    my_cgroup = proc_self_cgroup.split(':')[2].rstrip()
    cgroup_filepath = "/sys/fs/cgroup" + my_cgroup
    cgroup_filepath_parent = cgroup_filepath + "/.."

    # technically, also need to verify that parent's cgroup.subtree_control
    # contains memory, pid, cpu, but let's skip that for now
    return os.access(cgroup_filepath_parent, os.W_OK)


@pytest.mark.skipif(not CgroupIsWriteable(), reason="Test was not born in a writeable cgroup v2")
class TestCgroupOOM:
    def test_cgroup_oom(self, completed_test_job):
        assert completed_test_job.state.all_held()

    def test_cgroup_escape(self, escaping_test_job):
        # When the job exits, pidfile contains the pid of the sleep job
        with open('pidfile', 'r') as file:
            pid = file.read().rstrip()
        # Check that the pid does not exist in the system anymore -- that it has not escaped.
        assert not os.path.exists("/proc/" + pid)

    def test_cgroup_isolated(self, isolated_test_job):
        # When the job exits, this is a copy of the cgroup.procs file
        # from the cgroup of the job
        cgroup_procs = "cgroup.procs"
        assert os.path.exists(cgroup_procs)
        with open(cgroup_procs, 'r') as file:
            for count, pid in enumerate(file):
                pass
        count += 1 # Python has 0-indexing
        assert count == 1

    def test_cgroup_slice_scope(self, slice_scope_test_job):
        # When the job exits, cgroup contains the cgroup of the cat job
        with open('cgroup', 'r') as file:
            line        = file.read().rstrip()
            ultimate    = line.split('/')[-1]
            penultimate = line.split('/')[-2]
            assert ultimate.endswith(".scope")
            assert penultimate.endswith(".slice")
        assert True

    def test_cgroup_memory_detection(self, memory_detection_test_job):
        # When the job exits, detected_memory contains condor's idea of memory
        memory = 0
        with open('detected_memory', 'r') as file:
            memory = int(file.read().rstrip())
        assert(50 < memory and memory < 200) # Allow for rounding up to 128 Mb

    def test_cgroup_with_hook(self, hook_test_job):
        assert hook_test_job.state.all_held()

    def test_cgroup_with_retry_memory(self, retry_memory_job):
        assert retry_memory_job.state.all_complete()

    def test_retry_memory_without_units_fails(self, condor_with_units_error, retry_memory_no_units_job_hash):
        # Test that submitting retry_request_memory without units fails when SUBMIT_REQUEST_MISSING_UNITS=error
        with pytest.raises(Exception):
            condor_with_units_error.submit({**retry_memory_no_units_job_hash}, count=1)

    def test_retry_memory_max_without_units_fails(self, condor_with_units_error, retry_memory_max_no_units_job_hash):
        # Test that submitting retry_request_memory_max without units fails when SUBMIT_REQUEST_MISSING_UNITS=error
        with pytest.raises(Exception):
            condor_with_units_error.submit({**retry_memory_max_no_units_job_hash}, count=1)

    def test_retry_memory_with_units_succeeds(self, condor_with_units_error, retry_memory_with_units_job_hash):
        # Test that submitting retry_request_memory with units succeeds even when SUBMIT_REQUEST_MISSING_UNITS=error
        job = condor_with_units_error.submit({**retry_memory_with_units_job_hash}, count=1)
        assert job.wait(
            condition=ClusterState.all_complete,
            timeout=120,
            verbose=True,
            fail_condition=ClusterState.any_held,
        )

    def test_local_universe_oom(self, completed_local_test_job):
        # Test that a local universe job that exceeds memory is put on hold
        assert completed_local_test_job.state.all_held()



import os
import sys

from docutils import nodes
from docutils.parsers.rst import Directive

"""
Example of a Sphinx Directive.
This is useful for for working with large blocks of text. Leaving here as a reference.
Roles are more useful for inline text, which is what we need for tickets.

class Ticket(Directive):
    required_arguments = 1
    def run(self):
        ticket_url = "<a href=\"https://htcondor-wiki.cs.wisc.edu/index.cgi/tktview?tn=" + str(self.arguments[0]) + "\" target=\"blank\">link text</a>"
        paragraph_node = nodes.paragraph(text=ticket_url)
        return [paragraph_node]

def setup(app):
    app.add_directive("ticket", Ticket)

"""

# Custom role reference: https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/

def make_link_node(rawtext, app, type, slug, options):
    """Create a link to an HTCondor ticket.

    :param rawtext: Text being replaced with link node.
    :param app: Sphinx application context
    :param type: Link type (issue, changeset, etc.)
    :param slug: ID of the ticket to link to
    :param options: Options dictionary passed to role func.
    """
    base = "https://htcondor-wiki.cs.wisc.edu/index.cgi/tktview?tn="
    ref = base + slug
    # set_classes(options)
    node = nodes.reference(rawtext, "(Ticket #" + slug + ")", refuri=ref, **options)
    return node

def ticket_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    try:
        ticket_id = int(text)
        if ticket_id <= 1000:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'HTCondor ticket number must be a number greater than or equal to 1000; '
            '"%s" is invalid.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    app = inliner.document.settings.env.app
    node = make_link_node(rawtext, app, 'issue', str(ticket_id), options)
    return [node], []

def setup(app):
    app.add_role("ticket", ticket_role)




#!/usr/bin/env pytest

import classad2
import htcondor2

from ornithology import (
    action,
    Condor,
)


CASES = {
    "list_all_clusters": ["1", "2", "3"],
    "list_all_pairs": ["3.0", "3.1", "4.0"],
    "list_mixed": ["5.0", "6", "6.1"],

    "int": 7,
    "exprtree": classad2.ExprTree("ClusterID == 8"),
    "str_cluster": "9",
    "str_pair": "10.0",
    "str_expr": "ClusterID == 11 && ProcID == 1",
}


@action(params={name: name for name in CASES.keys()})
def case_name(request):
    return request.param


@action
def case(case_name):
    return CASES[case_name]


@action
def schedd(default_condor, path_to_sleep):
    description = {
        "executable":       path_to_sleep,
        "arguments":        60,
        "hold":             True,
        "log":              "$(ClusterID).$(ProcID).log",
    }

    schedd = default_condor.get_local_schedd();
    submit = htcondor2.Submit(description)

    for i in range(11):
        schedd.submit(submit, count=2)

    return schedd


class TestJobSpecHack:

    def test_invalid_type(self, schedd):
        try:
            schedd.edit( 7.5, 'invalid_type', 7.5 )
            assert False
        except TypeError:
            pass


    def test_invalid_list(self, schedd):
        try:
            schedd.edit( ["a", "b"], 'invalid_list', "a, b" )
            assert False
        except ValueError:
            pass


    def test_invalid_string(self, schedd):
        try:
            schedd.edit( "a!b", 'invalid_string', "a!b" )
            assert False
        except ValueError:
            pass


    # Note that we're not presently checking to make sure that we edit()ed
    # the _right_ jobs, but we could add that pretty easily.
    def test_valid_case(self, schedd, case_name, case):
        schedd.edit( case, case_name, '"valid"' )
        ads = schedd.query( projection=['ClusterID', 'ProcID', case_name] )
        assert len(ads) > 0
        relevant_ads = [ad for ad in ads if case_name in ad]
        assert len(relevant_ads) > 0
        assert all([ad[case_name] == "valid" for ad in relevant_ads])



#!/usr/bin/env pytest

# Test HTCondor offline ads functionality.
#
# HTCondor's collector supports two related features:
#
#   Offline ads – Machine ClassAds that are stored persistently on disk by
#   the collector's OfflineCollectorPlugin, even after the startd stops
#   updating them (e.g. because the machine is hibernating). They survive
#   a collector restart and are reloaded into the in-memory store on
#   start-up.  A machine ad becomes offline either by carrying Offline=true
#   when advertised (UPDATE_STARTD_AD) or by being sent via the
#   UPDATE_STARTD_AD_WITH_ACK command.
#
#   Absent ads – When ABSENT_REQUIREMENTS is configured and a machine ad
#   expires (its ClassAdLifetime elapses without a refresh), the collector
#   evaluates the expression against the expiring ad instead of deleting it.
#   If the expression is true the ad is marked Absent=true, given a fresh
#   ClassAdLifetime (ABSENT_EXPIRE_ADS_AFTER), and kept in the in-memory
#   store.  Normal (condor_status-style) queries automatically filter out
#   absent ads; they can be retrieved by including "Absent == true" in the
#   query constraint.

import logging
import time

import classad2 as classad
import htcondor2 as htcondor

from ornithology import (
    standup,
    action,
    Condor,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

# Fake address for synthetic machine ads.  We are not running a real startd,
# so we use a placeholder that satisfies the collector's hash-key requirements.
_FAKE_ADDR = "<127.0.0.1:9999?addrs=127.0.0.1-9999&noUDP>"


def _machine_ad(name, **extra):
    """Return a minimal Machine ClassAd suitable for advertising to a collector."""
    attrs = {
        "MyType": "Machine",
        "TargetType": "Job",
        "Name": name,
        "Machine": name,
        "MyAddress": _FAKE_ADDR,
        "StartdIpAddr": "127.0.0.1",
        "IsPytest": True,
    }
    attrs.update(extra)
    return classad.ClassAd(attrs)


# ──────────────────────────────────────────────────────────────────────────────
# Shared fixtures for offline-ad tests
# ──────────────────────────────────────────────────────────────────────────────

@standup
def condor_offline(test_dir):
    """Minimal HTCondor pool (collector + master) with persistent offline ad storage."""
    with Condor(
        local_dir=test_dir / "condor",
        config={
            "DAEMON_LIST": "COLLECTOR MASTER",
            "USE_SHARED_PORT": False,
            # Enable persistent offline ad storage.
            "COLLECTOR_PERSISTENT_AD_LOG": str(test_dir / "offline.ads"),
        },
    ) as condor:
        yield condor


@action
def offline_machine_name():
    return "offline-test-machine.example.com"


@action
def advertised_offline_ad(condor_offline, offline_machine_name):
    """Advertise a machine ad with Offline=true and wait for it to appear in the
    collector's in-memory store.  Returns the queried ClassAd, or None on timeout."""
    collector = condor_offline.get_local_collector()
    ad = _machine_ad(offline_machine_name, Offline=True)
    constraint = f'Name == "{offline_machine_name}" && Offline == true'
    advertised = False
    for _ in range(30):
        try:
            if not advertised:
                collector.advertise([ad], "UPDATE_STARTD_AD")
                advertised = True
            ads = collector.query(htcondor.AdTypes.Startd, constraint)
            if ads:
                return ads[0]
        except Exception:
            pass  # Collector may be momentarily busy under load.
        time.sleep(1)
    return None


# ──────────────────────────────────────────────────────────────────────────────
# TestOfflineAdBasic – the ad appears with the right attributes
# ──────────────────────────────────────────────────────────────────────────────

class TestOfflineAdBasic:
    """An ad advertised with Offline=true is stored in the collector and on disk."""

    def test_offline_ad_appears_in_collector(self, advertised_offline_ad):
        assert advertised_offline_ad is not None, (
            "Timed out waiting for offline ad to appear in collector"
        )

    def test_offline_attribute_is_true(self, advertised_offline_ad):
        assert advertised_offline_ad is not None
        assert advertised_offline_ad.get("Offline") is True

    def test_offline_ad_name_is_correct(self, advertised_offline_ad, offline_machine_name):
        assert advertised_offline_ad is not None
        assert advertised_offline_ad.get("Name") == offline_machine_name

    def test_persistent_log_file_created(self, advertised_offline_ad, test_dir):
        """The offline plugin must write a non-empty persistent log file."""
        assert advertised_offline_ad is not None
        log_path = test_dir / "offline.ads"
        assert log_path.exists(), "Persistent offline ad log was not created"
        assert log_path.stat().st_size > 0, "Persistent offline ad log is empty"

    def test_machine_name_in_persistent_log(self, advertised_offline_ad, offline_machine_name, test_dir):
        """The machine name must be written into the persistent log."""
        assert advertised_offline_ad is not None
        log_path = test_dir / "offline.ads"
        content = log_path.read_text(errors="replace")
        assert offline_machine_name in content, (
            f'"{offline_machine_name}" not found in persistent log {log_path}'
        )


# ──────────────────────────────────────────────────────────────────────────────
# TestOfflineAdPersistence – the ad survives a collector restart
# ──────────────────────────────────────────────────────────────────────────────

@action
def offline_ad_after_restart(condor_offline, advertised_offline_ad, offline_machine_name):
    """Restart the collector and wait for the offline ad to reappear.  The
    OfflineCollectorPlugin replays ads from the persistent log at start-up.

    We use condor_offline.status() inside the retry loop rather than caching a
    Collector object.  status() calls get_local_collector() inside a fresh
    use_config() context on every invocation, which calls htcondor.reload_config()
    and re-reads the COLLECTOR_ADDRESS_FILE.  This is necessary because the pool
    uses COLLECTOR_HOST=$(CONDOR_HOST):0 (a random port chosen at start-up), so
    after a restart the collector picks a NEW port and updates the address file.
    A cached Collector object would silently keep connecting to the dead old port.
    """
    assert advertised_offline_ad is not None, "Prerequisite: offline ad must have been advertised"

    cp = condor_offline.run_command(["condor_restart", "-collector"])
    assert cp.returncode == 0, f"condor_restart -collector failed: {cp.stderr}"

    # Poll using status() so that the collector address is re-resolved on every
    # attempt after the restart.
    constraint = f'Name == "{offline_machine_name}" && Offline == true'
    for _ in range(60):
        try:
            ads = condor_offline.status(
                ad_type=htcondor.AdTypes.Startd,
                constraint=constraint,
            )
            if ads:
                return ads[0]
        except Exception:
            pass  # Collector may be briefly unavailable during restart.
        time.sleep(1)
    return None


class TestOfflineAdPersistence:
    """Offline ads are reloaded from the persistent log after a collector restart."""

    def test_offline_ad_survives_restart(self, offline_ad_after_restart):
        assert offline_ad_after_restart is not None, (
            "Offline ad did not reappear in collector after restart"
        )

    def test_offline_attribute_preserved_after_restart(self, offline_ad_after_restart):
        assert offline_ad_after_restart is not None
        assert offline_ad_after_restart.get("Offline") is True


# ──────────────────────────────────────────────────────────────────────────────
# TestOfflineAdInvalidation – INVALIDATE removes the ad from memory and disk
# ──────────────────────────────────────────────────────────────────────────────

@action
def offline_ad_invalidated(condor_offline, advertised_offline_ad, offline_machine_name):
    """Send INVALIDATE_STARTD_ADS for the offline machine and wait for it to
    disappear from the collector.  Returns True when gone, False on timeout."""
    assert advertised_offline_ad is not None, "Prerequisite: offline ad must have been advertised"

    collector = condor_offline.get_local_collector()
    invalidate_ad = classad.ClassAd({
        "MyType": "Machine",
        "Name": offline_machine_name,
        "Machine": offline_machine_name,
        "MyAddress": _FAKE_ADDR,
        "StartdIpAddr": "127.0.0.1",
    })

    constraint = f'Name == "{offline_machine_name}"'
    invalidated = False
    for _ in range(30):
        try:
            if not invalidated:
                collector.advertise([invalidate_ad], "INVALIDATE_STARTD_ADS")
                invalidated = True
            ads = collector.query(htcondor.AdTypes.Startd, constraint)
            if not ads:
                return True
        except Exception:
            pass  # Collector may be momentarily busy under load.
        time.sleep(1)
    return False


class TestOfflineAdInvalidation:
    """Sending INVALIDATE_STARTD_ADS removes an offline ad from the collector
    and from the persistent log (the ad must not reappear after a restart)."""

    def test_offline_ad_gone_after_invalidate(self, offline_ad_invalidated):
        assert offline_ad_invalidated, (
            "Offline ad did not disappear from collector after INVALIDATE_STARTD_ADS"
        )

    def test_destroy_recorded_in_persistent_log(
        self, offline_ad_invalidated, offline_machine_name, test_dir
    ):
        """Invalidation must write a DestroyClassAd entry to the persistent log.

        The ClassAdCollection log is append-only: old CreateClassAd / SetAttribute
        entries for the machine remain in the file even after the ad is destroyed.
        The correct signal that the ad was invalidated is the presence of an op-102
        (DestroyClassAd) record whose key is the machine's offline key.

        The offline key is produced by AdNameHashKey::sprint() – which formats the
        name as "< name >" – followed by RemoveAllWhitespace(), giving "<name>".
        The on-disk record therefore looks like:
            102 <offline-test-machine.example.com>
        """
        assert offline_ad_invalidated
        log_path = test_dir / "offline.ads"
        assert log_path.exists(), "Persistent offline ad log was not created"
        content = log_path.read_text(errors="replace")
        # Op 102 = CondorLogOp_DestroyClassAd; key has whitespace removed.
        expected = f"102 <{offline_machine_name}>"
        assert expected in content, (
            f"Expected DestroyClassAd entry '{expected}' not found in persistent log"
        )


# ──────────────────────────────────────────────────────────────────────────────
# Absent-ad tests
# ──────────────────────────────────────────────────────────────────────────────

@standup
def condor_absent(test_dir):
    """Minimal HTCondor pool configured for absent-ad testing.

    CLASSAD_LIFETIME is set to 15 seconds so machine ads expire quickly.
    ABSENT_REQUIREMENTS = True means every expiring machine ad becomes absent
    rather than being deleted.  ABSENT_EXPIRE_ADS_AFTER is set to a full day
    so absent ads stay around for the duration of the test.
    """
    with Condor(
        local_dir=test_dir / "condor",
        config={
            "DAEMON_LIST": "COLLECTOR MASTER",
            "USE_SHARED_PORT": False,
            "COLLECTOR_PERSISTENT_AD_LOG": str(test_dir / "absent.ads"),
            # Short ad lifetime so the synthetic ad expires quickly.
            "CLASSAD_LIFETIME": "15",
            # Mark every expiring machine ad as absent instead of deleting it.
            "ABSENT_REQUIREMENTS": "True",
            # Keep absent ads alive for the test.
            "ABSENT_EXPIRE_ADS_AFTER": "86400",
        },
    ) as condor:
        yield condor


@action
def absent_machine_name():
    return "absent-test-machine.example.com"


@action
def advertise_soon_expiring_ad(condor_absent, absent_machine_name):
    """Advertise a machine ad with ClassAdLifetime=5 seconds.

    The collector will expire this ad within seconds and, because
    ABSENT_REQUIREMENTS=True, will promote it to an absent ad.
    Returns True when the initial ad appears, False on timeout.
    """
    collector = condor_absent.get_local_collector()
    ad = _machine_ad(absent_machine_name, ClassAdLifetime=5)
    constraint = f'Name == "{absent_machine_name}"'
    advertised = False
    # Retry loop: the collector may not be ready to accept connections yet,
    # especially under load.  Re-advertise on failure so that the short
    # ClassAdLifetime doesn't expire before the ad is accepted.
    for _ in range(30):
        try:
            if not advertised:
                collector.advertise([ad], "UPDATE_STARTD_AD")
                advertised = True
            ads = collector.query(htcondor.AdTypes.Startd, constraint)
            if ads:
                return True
        except Exception:
            pass  # Collector may be momentarily busy under load.
        time.sleep(1)
    return False


@action
def absent_ad(condor_absent, absent_machine_name, advertise_soon_expiring_ad):
    """Wait for the expiring ad to be promoted to Absent=true.

    The collector's house-keeper expires the ad after ClassAdLifetime seconds
    and calls OfflineCollectorPlugin::expire(), which evaluates
    ABSENT_REQUIREMENTS.  If true the ad is kept with Absent=true.
    Returns the absent ClassAd, or None on timeout.
    """
    assert advertise_soon_expiring_ad, "Prerequisite: initial ad must have appeared in collector"
    collector = condor_absent.get_local_collector()
    constraint = f'Name == "{absent_machine_name}" && Absent == true'
    for _ in range(60):
        ads = collector.query(htcondor.AdTypes.Startd, constraint)
        if ads:
            return ads[0]
        time.sleep(2)
    return None


class TestAbsentAds:
    """When ABSENT_REQUIREMENTS matches an expiring machine ad, the collector
    retains it with Absent=true instead of deleting it."""

    def test_absent_ad_appears(self, absent_ad):
        assert absent_ad is not None, (
            "Timed out waiting for machine ad to be promoted to absent"
        )

    def test_absent_attribute_is_true(self, absent_ad):
        assert absent_ad is not None
        assert absent_ad.get("Absent") is True

    def test_absent_ad_name_is_correct(self, absent_ad, absent_machine_name):
        assert absent_ad is not None
        assert absent_ad.get("Name") == absent_machine_name

    def test_absent_ad_not_in_normal_query(self, condor_absent, absent_machine_name, absent_ad):
        """Absent ads must be invisible to ordinary queries.

        When ABSENT_REQUIREMENTS is configured the collector automatically
        filters out absent ads from any query that does not explicitly
        reference the Absent attribute.  This mimics the behaviour of
        condor_status without the -absent flag.
        """
        assert absent_ad is not None, "Prerequisite: absent ad must exist"
        collector = condor_absent.get_local_collector()
        # Plain constraint – does not mention Absent.
        ads = collector.query(
            htcondor.AdTypes.Startd,
            f'Name == "{absent_machine_name}"',
        )
        absent_in_results = [a for a in ads if a.get("Absent") is True]
        assert len(absent_in_results) == 0, (
            "Absent ad appeared in a normal (no-absent-constraint) query"
        )

    def test_absent_ad_visible_with_absent_constraint(
        self, condor_absent, absent_machine_name, absent_ad
    ):
        """Absent ads are returned when the query explicitly requests Absent=true.

        Including Absent in the constraint disables the server-side absent
        filter, so the absent ad is returned.  This mimics condor_status -absent.
        """
        assert absent_ad is not None, "Prerequisite: absent ad must exist"
        collector = condor_absent.get_local_collector()
        ads = collector.query(
            htcondor.AdTypes.Startd,
            f'Name == "{absent_machine_name}" && Absent == true',
        )
        assert len(ads) == 1, (
            f"Expected 1 absent ad for {absent_machine_name}, got {len(ads)}"
        )

    def test_absent_ad_in_persistent_log(self, absent_ad, absent_machine_name, test_dir):
        """Absent ads are written to the persistent log so they survive restarts."""
        assert absent_ad is not None, "Prerequisite: absent ad must exist"
        log_path = test_dir / "absent.ads"
        assert log_path.exists(), "Persistent absent ad log was not created"
        content = log_path.read_text(errors="replace")
        assert absent_machine_name in content, (
            f'"{absent_machine_name}" not found in persistent log {log_path}'
        )



#!/usr/bin/env pytest

import pytest
import subprocess

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from ornithology import (
    Condor,
    action,
    ClusterState,
    format_script,
    write_file,
    JobStatus,
)

import os
from pathlib import Path
import shutil

import htcondor2


#
# This is based on `test_cif.py`; these test aren't included there because
# (a) that test already takes close to two minutes to run, and
# (b) the helper functions in this test will have to be a little different.
#

#
# The basic idea is similar: we submit some jobs which stitch together the
# input files to make an output file with known contents.  For these tests,
# the jobs will `cat` _all_ of the input files: this not only confirms that
# we got the _correct_ input files, but that we did _not_ get any of the
# others.  (Unless they're empty, which we'll check for with an `ls` as part
# of the output, I guess.)
#
# In particular, we'll submit two sets of jobs: set 1 will have catalogs A
# and B, and set 2 will have catalogs A and C.  The catalogs will have
# disjoint file membership, and each file will have unique contents.
#
# As in `test_cif.py`, we'll also be verifying that we transferred the
# catalogs the correct number of times.  This leads to two test cases
# for the preceeding scenario: one with two invocations of condor_submit,
# and one with one invocation of DAGMan.  In the former, there must be
# four transfers: A twice (once for each cluster) and B and C once each
# (once for each cluster).  In the latter, A is only transferred once.
#

#
# It remains to be seen how we can test the requirement that catalogs
# named by different submmiters (the Owner attribute, not the misnamed
# Submitter attribute that's actually for accounting) are different.
#

@action
def the_cs_local_dir(test_dir):
    return test_dir / "cs.d"


@action
def the_cs_user_dir(the_cs_local_dir):
    the_cs_user_dir = the_cs_local_dir / "user.d"
    the_cs_user_dir.mkdir(exist_ok=True)
    return the_cs_user_dir


@action
def the_cs_lock_dir(the_cs_local_dir):
    return the_cs_local_dir / "lock.d"


@action
def the_cs_condor(the_cs_local_dir, the_cs_lock_dir):
    with Condor(
        local_dir=the_cs_local_dir,
        config={
            "FORBID_COMMON_FILE_TRANSFER":          "TRUE",

            "STARTER_DEBUG":            "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":             "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":                     the_cs_lock_dir.as_posix(),
            "NUM_CPUS":                 4,
            "STARTER_NESTED_SCRATCH":   True,
        },
    ) as the_cs_condor:
        yield the_cs_condor


@action
def the_cs_job_script(test_dir):
    script_path = test_dir / "cs_job_script"

    script_text = format_script(
    """
        #!/bin/sh

        ls *.txt
        cat *.txt

        while true; do
            sleep 1
            if [ -e $1 ]; then
                exit 0
            fi
        done

        exit 1
    """
    )

    write_file(script_path, script_text)
    return script_path


@action
def completed_cs_jobs(the_cs_condor, the_cs_user_dir, the_cs_job_script):
    os.chdir(the_cs_user_dir)

    # We did all the complicated nested-directory stuff
    # in `test_cif.py`, so this is just a pile of files.
    for f in ("A1", "A2", "B1", "B2", "C1", "C2"):
        (the_cs_user_dir / f"{f}.txt").write_text(f"{f}\n")

    kill_file = (the_cs_user_dir / "kill-cs-$(ClusterID).$(ProcID)")
    job_description = {
        "universe":                 "vanilla",
        "executable":               the_cs_job_script.as_posix(),
        "arguments":                kill_file.as_posix(),

        "log":                      "cs_job.log.$(CLUSTER)",
        "output":                   "cs_job.output.$(CLUSTER).$(PROCESS)",
        "error":                    "cs_job.error.$(CLUSTER).$(PROCESS)",

        "request_cpus":             1,
        "request_memory":           1,

        "MY._x_catalog_A":                  '"A1.txt, A2.txt"',

        "should_transfer_files":    True,

        "leave_in_queue":           True,
    }

    job_description_a = {
        ** job_description,
        "MY._x_common_input_catalogs":      '"A, B"',
        "MY._x_catalog_B":                  '"B1.txt, B2.txt"',
    }

    job_description_b = {
        ** job_description,
        "MY._x_common_input_catalogs":      '"A, C"',
        "MY._x_catalog_C":                  '"C1.txt, C2.txt"',
    }


    # Submit two jobs of each type.
    job_handle_a = the_cs_condor.submit(
        description=job_description_a,
        count=2
    )

    job_handle_b = the_cs_condor.submit(
        description=job_description_b,
        count=2
    )


    # Wait for them all to start.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_running,
        fail_condition=ClusterState.any_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.running_exactly(2),
        fail_condition=ClusterState.any_terminal
    )


    # Touch their kill files.
    (the_cs_user_dir / f"kill-cs-{job_handle_a.clusterid}.0").touch(exist_ok=True)
    (the_cs_user_dir / f"kill-cs-{job_handle_a.clusterid}.1").touch(exist_ok=True)
    (the_cs_user_dir / f"kill-cs-{job_handle_b.clusterid}.0").touch(exist_ok=True)
    (the_cs_user_dir / f"kill-cs-{job_handle_b.clusterid}.1").touch(exist_ok=True)


    # Wait for them to finish.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )


    return job_handle_a, job_handle_b


# ---- (fake) DAGMan ------------------------------------------------------


@action
def the_dagman_local_dir(test_dir):
    return test_dir / "dm.d"


@action
def the_dagman_user_dir(the_dagman_local_dir):
    the_dagman_user_dir = the_dagman_local_dir / "user.d"
    the_dagman_user_dir.mkdir(exist_ok=True)
    return the_dagman_user_dir


@action
def the_dagman_lock_dir(the_dagman_local_dir):
    return the_dagman_local_dir / "lock.d"


@action
def the_dagman_condor(the_dagman_local_dir, the_dagman_lock_dir):
    with Condor(
        local_dir=the_dagman_local_dir,
        config={
            "FORBID_COMMON_FILE_TRANSFER":          "TRUE",

            "STARTER_DEBUG":            "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":             "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":                     the_dagman_lock_dir.as_posix(),
            "NUM_CPUS":                 4,
            "STARTER_NESTED_SCRATCH":   True,
        },
    ) as the_dagman_condor:
        yield the_dagman_condor


@action
def completed_dagman_jobs(the_dagman_condor, the_dagman_user_dir, the_cs_job_script):
    os.chdir(the_dagman_user_dir)

    # We did all the complicated nested-directory stuff
    # in `test_cif.py`, so this is just a pile of files.
    for f in ("A1", "A2", "B1", "B2", "C1", "C2"):
        (the_dagman_user_dir / f"{f}.txt").write_text(f"{f}\n")

    kill_file = (the_dagman_user_dir / "kill-dm-$(ClusterID).$(ProcID)")
    job_description = {
        "MY.DAGManJobID":           "20",

        "universe":                 "vanilla",
        "executable":               the_cs_job_script.as_posix(),
        "arguments":                kill_file.as_posix(),

        "log":                      "cs_job.log.$(CLUSTER)",
        "output":                   "cs_job.output.$(CLUSTER).$(PROCESS)",
        "error":                    "cs_job.error.$(CLUSTER).$(PROCESS)",

        "request_cpus":             1,
        "request_memory":           1,

        "MY._x_catalog_A":          '"A1.txt, A2.txt"',

        "should_transfer_files":    True,

        "leave_in_queue":           True,
    }

    # FIXME: We should test that:
    #
    # 1.  A, B and A, C with *B != *C produce different output.
    # 2.  A, B and A, B with *B != *C produce different output.
    # 3.  A, B and A, C with *B == *C produce the same output (but
    #     transfer common files twice).
    # 4.  A, B and A, B with *B == *B produce the same output (but
    #     transer common files once).
    job_description_a = {
        ** job_description,
        "MY._x_common_input_catalogs":      '"A, B"',
        "MY._x_catalog_B":                  '"B1.txt, B2.txt"',
    }

    job_description_b = {
        ** job_description,
        "MY._x_common_input_catalogs":      '"A, B"',
        "MY._x_catalog_B":                  '"C1.txt, C2.txt"',
    }


    # Submit two jobs of each type.
    job_handle_a = the_dagman_condor.submit(
        description=job_description_a,
        count=2
    )

    job_handle_b = the_dagman_condor.submit(
        description=job_description_b,
        count=2
    )


    # Wait for them all to start.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_running,
        fail_condition=ClusterState.any_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.running_exactly(2),
        fail_condition=ClusterState.any_terminal
    )


    # Touch their kill files.
    (the_dagman_user_dir / f"kill-dm-{job_handle_a.clusterid}.0").touch(exist_ok=True)
    (the_dagman_user_dir / f"kill-dm-{job_handle_a.clusterid}.1").touch(exist_ok=True)
    (the_dagman_user_dir / f"kill-dm-{job_handle_b.clusterid}.0").touch(exist_ok=True)
    (the_dagman_user_dir / f"kill-dm-{job_handle_b.clusterid}.1").touch(exist_ok=True)


    # Wait for them to finish.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )


    return job_handle_a, job_handle_b


# ---- common containers ------------------------------------------------------


@action
def the_container_image(test_dir, pytestconfig):
    # This is a gross hack.
    ctest_path = test_dir / ".." / "busybox.sif"
    if ctest_path.exists():
        return ctest_path
    else:
        return Path(pytestconfig.rootdir) / "busybox.sif"


@action
def the_container_local_dir(test_dir):
    return test_dir / "cc.d"


@action
def the_container_user_dir(the_container_local_dir):
    the_container_user_dir = the_container_local_dir / "user.d"
    the_container_user_dir.mkdir(exist_ok=True)
    return the_container_user_dir


@action
def the_container_kill_dir(test_dir):
    the_container_kill_dir = test_dir / "kill.d"
    the_container_kill_dir.mkdir(exist_ok=True)
    return the_container_kill_dir


@action
def the_container_lock_dir(the_container_local_dir):
    return the_container_local_dir / "lock.d"


@action
def the_container_condor(the_container_local_dir, the_container_lock_dir, the_container_kill_dir):
    with Condor(
        local_dir=the_container_local_dir,
        config={
            "FORBID_COMMON_FILE_TRANSFER":          "TRUE",

            "STARTER_DEBUG":            "D_CATEGORY D_SUB_SECOND D_PID D_ACCOUNTANT",
            "SHADOW_DEBUG":             "D_CATEGORY D_SUB_SECOND D_PID D_TEST",
            "LOCK":                     the_container_lock_dir.as_posix(),
            "NUM_CPUS":                 4,
            "STARTER_NESTED_SCRATCH":   True,
            "SINGULARITY_TEST_SANDBOX_TIMEOUT":              "8",
            "SINGULARITY":              "/usr/bin/singularity",
            "SINGULARITY_BIND_EXPR":    f'"{the_container_kill_dir.as_posix()}:{the_container_kill_dir.as_posix()}"',
            "CONTAINER_IMAGES_COMMON_BY_DEFAULT":   True,
        },
    ) as the_container_condor:
        yield the_container_condor


@action
def completed_container_jobs(the_container_condor, the_container_user_dir, the_cs_job_script, the_container_image, the_container_kill_dir):
    os.chdir(the_container_user_dir)

    # We did all the complicated nested-directory stuff
    # in `test_cif.py`, so this is just a pile of files.
    for f in ("A1", "A2"):
        (the_container_user_dir / f"{f}.txt").write_text(f"{f}\n")

    kill_file = the_container_kill_dir / "kill-cc-$(ClusterID).$(ProcID)"
    job_description = {
        "universe":                 "vanilla",
        "executable":               the_cs_job_script.as_posix(),
        "arguments":                kill_file.as_posix(),
        "container_image":          the_container_image.as_posix(),
        "transfer_executable":      True,

        "log":                      "cc_job.log.$(CLUSTER)",
        "output":                   "cc_job.output.$(CLUSTER).$(PROCESS)",
        "error":                    "cc_job.error.$(CLUSTER).$(PROCESS)",

        "request_cpus":             1,
        "request_memory":           1,

        "MY.CommonInputFiles":      '"A1.txt, A2.txt"',

        "should_transfer_files":    True,

        "leave_in_queue":           True,
    }

    job_description_a = {
        ** job_description,
    }

    job_description_b = {
        ** job_description,
        "container_is_common": "False",
    }


    # Submit two jobs of each type.
    job_handle_a = the_container_condor.submit(
        description=job_description_a,
        count=2
    )

    job_handle_b = the_container_condor.submit(
        description=job_description_b,
        count=2
    )


    # Wait for them all to start.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_running,
        fail_condition=ClusterState.any_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.running_exactly(2),
        fail_condition=ClusterState.any_terminal
    )


    # Touch their kill files.
    (the_container_kill_dir / f"kill-cc-{job_handle_a.clusterid}.0").touch(exist_ok=True)
    (the_container_kill_dir / f"kill-cc-{job_handle_a.clusterid}.1").touch(exist_ok=True)
    (the_container_kill_dir / f"kill-cc-{job_handle_b.clusterid}.0").touch(exist_ok=True)
    (the_container_kill_dir / f"kill-cc-{job_handle_b.clusterid}.1").touch(exist_ok=True)


    # Wait for them to finish.
    assert job_handle_a.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )
    assert job_handle_b.wait(
        timeout=60,
        condition=ClusterState.all_terminal
    )


    return job_handle_a, job_handle_b


# ---- assertion helpers ------------------------------------------------------


def output_is_as_expected(completed_cif_job, the_expected_output):
    ads = completed_cif_job.query( projection=['Out'] )

    for ad in ads:
        out = Path(ad['Out'])
        text = out.read_text()
        assert text == the_expected_output


def count_shadow_log_lines(the_condor, the_phrase):
    shadow_log = the_condor.shadow_log.open()
    lines = list(filter(
        lambda line: the_phrase in line,
        shadow_log.read(),
    ))
    return len(lines)


def shadow_log_is_as_expected(the_condor, count, cf_xfers, cf_waits):
    staging_commands_sent = count_shadow_log_lines(
        the_condor, "StageCommonFiles"
    )
    assert staging_commands_sent == count

    successful_staging_commands = count_shadow_log_lines(
        the_condor, "Staging successful"
    )
    assert successful_staging_commands == count

    keyfile_touches = count_shadow_log_lines(
        the_condor, "Elected producer touch"
    )
    assert keyfile_touches == count

    job_evictions = count_shadow_log_lines(
        the_condor, "is being evicted from"
    )
    assert job_evictions == 0

    common_transfer_begins = count_shadow_log_lines(
        the_condor, "Starting common files transfer."
    )
    assert common_transfer_begins == cf_xfers

    common_transfer_ends = count_shadow_log_lines(
        the_condor, "Finished common files transfer: success."
    )
    assert common_transfer_ends == cf_xfers

    if cf_waits is not None:
        common_transfer_waits = count_shadow_log_lines(
            the_condor, "Waiting for common files to be transferred"
        )
        assert common_transfer_waits == cf_waits


def lock_dir_is_clean(the_lock_dir):
    syndicate_dir = the_lock_dir / "syndicate"

    if syndicate_dir.exists():
        files = list(syndicate_dir.iterdir())
        assert len(files) == 0


# ---- Singularity checks -----------------------------------------------------
# All stolen from `test_singularity_sif.py`, and which should probably be made
# Ornithology fixtures.

def SingularityIsWorthy():
    result = subprocess.run("singularity --version", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output = result.stdout.decode('utf-8')

    logger.debug(output)
    if "apptainer" in output:
        return True

    if "3." in output:
        return True

    return False


def SingularityIsWorking():
    result = subprocess.run("singularity exec -B/bin:/bin -B/lib:/lib -B/lib64:/lib64 -B/usr:/usr busybox.sif /bin/ls /", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output = result.stdout.decode('utf-8')

    logger.debug(output)

    if result.returncode == 0:
        return True
    else:
        return False


# For the test to work, we need user namespaces to be working
# and enough of them.  This is a race, but better to try
# to test first.
def UserNamespacesFunctional():
    result = subprocess.run(["unshare", "-U", "/bin/sh", "-c", "exit 7"])
    if result.returncode == 7:
        print("unshare seems to work correctly, proceeding with test\n")
        return True
    else:
        print("unshare command failed, test cannot work, skipping test\n")
        return False


# ---- Tests ------------------------------------------------------------------


class TestCIFCatalogs:

    def test_condor_submit(self, the_cs_lock_dir, the_cs_condor, completed_cs_jobs):
        output_is_as_expected(
            completed_cs_jobs[0],
            "A1.txt\nA2.txt\nB1.txt\nB2.txt\nA1\nA2\nB1\nB2\n"
        )
        output_is_as_expected(
            completed_cs_jobs[1],
            "A1.txt\nA2.txt\nC1.txt\nC2.txt\nA1\nA2\nC1\nC2\n"
        )
        # Specifically, A should be transferred twice, B once, and C once.
        # If we later care about how many transfers waited, see `test_cif.py`.
        shadow_log_is_as_expected(the_cs_condor, 0, 0, None)
        lock_dir_is_clean(the_cs_lock_dir)


    def test_dagman(self, the_dagman_lock_dir, the_dagman_condor, completed_dagman_jobs):
        output_is_as_expected(
            completed_dagman_jobs[0],
            "A1.txt\nA2.txt\nB1.txt\nB2.txt\nA1\nA2\nB1\nB2\n"
        )
        output_is_as_expected(
            completed_dagman_jobs[1],
            "A1.txt\nA2.txt\nC1.txt\nC2.txt\nA1\nA2\nC1\nC2\n"
        )
        # Specifically, A should be transferred once, B once, and C once.
        # If we later care about how many transfers waited, see `test_cif.py`.
        shadow_log_is_as_expected(the_dagman_condor, 0, 0, None)
        lock_dir_is_clean(the_dagman_lock_dir)


    @pytest.mark.skipif(not SingularityIsWorthy(), reason="No worthy Singularity/Apptainer found")
    @pytest.mark.skipif(not UserNamespacesFunctional(), reason="User namespaces not working -- some limit hit?")
    @pytest.mark.skipif(not SingularityIsWorking(), reason="Singularity doesn't seem to be working")
    def test_container(self, the_container_lock_dir, the_container_condor, completed_container_jobs):
        output_is_as_expected(
            completed_container_jobs[0],
            "A1.txt\nA2.txt\nA1\nA2\n"
        )
        output_is_as_expected(
            completed_container_jobs[1],
            "A1.txt\nA2.txt\nA1\nA2\n"
        )
        # Specifically, the container should be transferred once,
        # and A should be transferred twice.  (The container will be
        # transferred three times, but the last two won't be common
        # transfers, which is what we're counting here.)
        shadow_log_is_as_expected(the_container_condor, 0, 0, None)
        lock_dir_is_clean(the_container_lock_dir)



import sys
from pathlib import Path
import datetime
from collections import defaultdict

import htcondor2 as htcondor
import classad2 as classad

from htcondor_cli.noun import Noun
from htcondor_cli.verb import Verb


def compactify_job_list(jobs):
    if jobs is None:
        return ""

    list = ""
    for clusterID in sorted(jobs.keys()):
        ordered = sorted(jobs[clusterID])

        # This whole thing needs some help.
        if len(ordered) == 1:
            list += f"{clusterID}.{ordered[0]}; "
        elif len(ordered) == 2:
            list += f"{clusterID}.{ordered[0]},{ordered[1]}; "
        else:
            list += f"{clusterID}."

            last = None
            first = None
            index = 0

            while index < len(ordered):
                if index + 1 == len(ordered):
                    last = ordered[index]
                    next = None
                else:
                    next = ordered[index + 1]

                if ordered[index] + 1 == next:
                    if first is None:
                        first = ordered[index]
                    last = ordered[index + 1]
                else:
                    if first is None:
                        list += f"{ordered[index]},"
                    else:
                        list += f"[{first}-{last}],"
                        first = None
                        last = None
                index += 1

            list = list[:-1]
            list += "; "

    return list[:-2]


class ListAll(Verb):
    def __init__(self, logger, **options):
        # Grovel around in SEC_CREDENTIAL_DIRECTORY_OAUTH.
        credentials_dir_path = htcondor.param.get("SEC_CREDENTIAL_DIRECTORY_OAUTH")
        if credentials_dir_path is None:
            return
        credentials_dir = Path(credentials_dir_path)


        credentials_by_user = defaultdict(dict)
        try:
            for entry in credentials_dir.iterdir():
                if not entry.is_dir():
                    continue

                for file in entry.iterdir():
                    if file.suffix != ".top" and file.suffix != ".use":
                        continue
                    credentials_by_user[entry.name][file.stem] = {
                        'refresh':  int(file.stat().st_mtime),
                        'jobs':     defaultdict(list),
                    }
        except PermissionError:
            print("This command is meant to be run by the access point administrator.  (You don't have permission to read the credentials directory.)")
            return

        # We're assuming that the unqualified user names on disk are
        # the owner names in the queue, but that's not necessarily
        # true; at some point, we'll have to update the credd/credmon.
        schedd = htcondor.Schedd()
        results = schedd.query(
            constraint=f'OAuthServicesNeeded =!= undefined',
            projection=['Owner', 'OAuthServicesNeeded', 'ClusterID', 'ProcID']
        )
        for result in results:
            clusterID = result['ClusterID']
            procID = result['ProcID']
            owner = result['Owner']
            service_names = result['OAuthServicesNeeded']
            for service_name in service_names.split(','):
                service_name = service_name.replace('*', '_')
                # In the wild, not all values in OAuthServicesNeeded are correctly
                # constructed (e.g., are space-separated), and the lookup may fail.
                credentials_by_service = credentials_by_user[owner].get(service_name)
                if credentials_by_service is not None:
                    credentials_by_service['jobs'][clusterID].append(procID)

        if len(credentials_by_user.keys()) == 0:
            print(f">> no OAuth credentials found")
            return

        longest_user = 4
        for user in credentials_by_user.keys():
            if len(user) > longest_user:
                longest_user = len(user)

        longest_file = 4
        for entry in credentials_by_user.values():
            for file in entry.keys():
                if len(file) + 4 > longest_file:
                    longest_file = len(file) + 4

        print(f">> Found the following OAuth2 credentials:")
        print(f"   {'User':<{longest_user}}  {'Service':<18}  {'Handle':<18}  {'Last Refreshed':>19}  {'File':<{longest_file}}  Jobs")

        for user in sorted(credentials_by_user.keys()):
            for name in credentials_by_user[user].keys():
                entry = credentials_by_user[user][name]
                (service, _, handle) = name.partition('_')
                refresh = str(datetime.datetime.fromtimestamp(entry['refresh']))
                jobs = compactify_job_list(entry.get('jobs'))

                file = f"{name}.use"

                print(f"   {user:<{longest_user}}  {service:<18}  {handle:<18}  {refresh:>19}  {file:<{longest_file}}  {jobs}")


class List(Verb):
    # For now, this only works for the authenticated identity and the
    # various different local daemons (depending on the command).
    def __init__(self, logger, **options):
        user = None
        credd = htcondor.Credd()

        # Check for a password credential.
        has_windows_password = False
        windows_password_time = None
        try:
            # This ignores the specified daemon and asks the schedd directly;
            # the schedd doesn't even register the corresponding command except
            # on Windows, so we get an exception instead of an answer.  Sigh.
            has_windows_password = credd.query_password(user)
            windows_password_time = credd.query_user_cred(htcondor.CredTypes.Password, user)
        except htcondor.HTCondorException:
            pass
        if windows_password_time is not None:
            date = datetime.datetime.fromtimestamp(windows_password_time)
            print(f">> Windows password timestamp {date}")
        else:
            print(f">> no Windows password found");

        # Check for Kerberos credential.
        kerberos_time = None
        try:
            # This throws an exception if the Kerberos credential directory
            # is misconfigured, which is super not-helpful, because if it's
            # configured but not active, the schedd blocks for ten minutes
            # on start-up.
            kerberos_time = credd.query_user_cred(htcondor.CredTypes.Kerberos, user)
        except htcondor.HTCondorException:
            pass
        if kerberos_time is not None:
            date = datetime.datetime.fromtimestamp(kerberos_time)
            print(f">> Kerberos timestamp {date}")
        else:
            print(f">> no Kerberos credential found")

        # Check for OAuth2 credentials.
        oauth_classad_string = credd.query_user_service_cred(htcondor.CredTypes.OAuth, None, None, user)

        if oauth_classad_string is None:
            print(f">> no OAuth credentials found")
            return

        ad = classad.parseOne(oauth_classad_string)
        # If we're talking to an older server, this could be None; we
        # handle that in the formatting step, below.
        fully_qualified_user = ad.get("fully_qualified_user")

        names = {}
        for key in ad.keys():
            name = None
            if key.endswith(".top"):
                name = key[:-4]
            elif key.endswith(".use"):
                name = key[:-4]

            if name is not None:
                names[name] = ad[key]

        fqu_string = ""
        if fully_qualified_user is not None:
            fqu_string = f" for '{fully_qualified_user}'"
        if len(names.keys()) == 0:
            print(f">> no OAuth credentials found{fqu_string}")
            return
        fqu_string += ':'

        #
        # Let's ask the queue for jobs that use these credentials.
        #
        jobs_by_name = defaultdict(lambda: defaultdict(list))
        if fully_qualified_user is not None:
            schedd = htcondor.Schedd()
            results = schedd.query(
                constraint=f'user == "{fully_qualified_user}" && OAuthServicesNeeded =!= undefined',
                projection=['OAuthServicesNeeded', 'ClusterID', 'ProcID']
            )
            for result in results:
                clusterID = result['ClusterID']
                procID = result['ProcID']
                service_names = result['OAuthServicesNeeded']
                for service_name in service_names.split(','):
                    service_name = service_name.replace('*', '_')
                    jobs_by_name[service_name][clusterID].append(procID)


        # Convert 'File' to a fixed-width column.
        longest_name = 4
        for name in names:
            if len(name) + 4 > longest_name:
                longest_name = len(name) + 4

        print(f">> Found OAuth2 credentials{fqu_string}")
        print(f"   {'Service':<18}  {'Handle':<18}  {'Last Refreshed':>19}  {'File':<{longest_name}}  Jobs")
        for name in names:
            (service, _, handle) = name.partition('_')
            date = datetime.datetime.fromtimestamp(names[name])
            jobs = compactify_job_list(jobs_by_name.get(name))
            name = f"{name}.use"
            print(f"   {service:<18}  {handle:<18}  {str(date):>19}  {name:<{longest_name}}  {jobs}")


class Get(Verb):
    def _handle_get_password(** options):
        raise NotImplementedException


    def _handle_get_kerberos(** options):
        raise NotImplementedException


    def _handle_get_oauth2(*, service, handle, ** options):
        user = None
        credd = htcondor.Credd()
        credential = credd.get_oauth2_credential(service, handle, user)
        print(credential)


    choices = {
        "password":     _handle_get_password,
        "kerberos":     _handle_get_kerberos,
        "oauth2":       _handle_get_oauth2,
    }


    options = {
        "type": {
            "args":         ("type",),
            "metavar":      "type",
            "choices":      choices.keys(),
            "help":         "The credential type: password, kerberos, or oauth2",
        },
        "service": {
            "args":         ("--service",),
            "metavar":      "service",
            "help":         "(OAuth2)  Service name, if not from the file",
        },
        "handle": {
            "args":         ("--handle",),
            "metavar":      "handle",
            "help":         "(OAuth2)  Handle name, if not from the file",
        },
    }


    def __init__(self, logger, **options):
        self.choices[options['type']](** options)


class Add(Verb):
    def _handle_add_password(*, credential_file, ** options):
        user = None
        credd = htcondor.Credd()
        contents = Path(credential_file).read_text()

        # This can't work except for on Windows...
        credd.add_user_cred(htcondor.CredTypes.Password, contents, user)


    def _handle_add_kerberos(*, credential_file, ** options):
        user = None
        credd = htcondor.Credd()
        contents = Path(credential_file).read_text()

        credd.add_user_cred(htcondor.CredTypes.Kerberos, contents, user)


    def _handle_add_oauth2(*, credential_file, service, handle, ** options):
        user = None
        credd = htcondor.Credd()
        contents = Path(credential_file).read_text()

        if service is None and handle is None:
            (service, _, handle) = Path(credential_file).name.partition('_')

        credd.add_user_service_cred(htcondor.CredTypes.OAuth, contents, service, handle, user)


    choices = {
        "password":     _handle_add_password,
        "kerberos":     _handle_add_kerberos,
        "oauth2":       _handle_add_oauth2,
    }


    options = {
        "type": {
            "args":         ("type",),
            "metavar":      "type",
            "choices":      choices.keys(),
            "help":         "The credential type: password, kerberos, or oauth2",
        },
        "credential-file": {
            "args":         ("credential_file",),
            "metavar":      "credential-file",
            "help":         "Path to a file storing the credential",
        },
        "service": {
            "args":         ("--service",),
            "metavar":      "service",
            "help":         "(OAuth2)  Service name, if not from the file",
        },
        "handle": {
            "args":         ("--handle",),
            "metavar":      "handle",
            "help":         "(OAuth2)  Handle name, if not from the file",
        },
    }


    def __init__(self, logger, **options):
        self.choices[options['type']](** options)


class Remove(Verb):
    def _handle_delete_password(** options):
        user = None
        credd = htcondor.Credd()

        # This can't work except for on Windows...
        credd.delete_user_cred(htcondor.CredTypes.Password,user)


    def _handle_delete_kerberos(** options):
        user = None
        credd = htcondor.Credd()

        credd.delete_user_cred(htcondor.CredTypes.Kerberos, user)


    def _handle_delete_oauth2(*, service, handle, ** options):
        user = None
        credd = htcondor.Credd()

        credd.delete_user_service_cred(htcondor.CredTypes.OAuth, service, handle, user)


    choices = {
        "password":     _handle_delete_password,
        "kerberos":     _handle_delete_kerberos,
        "oauth2":       _handle_delete_oauth2,
    }


    options = {
        "type": {
            "args":         ("type",),
            "metavar":      "type",
            "choices":      choices.keys(),
            "help":         "The credential type: password, kerberos, or oauth2",
        },
        "service": {
            "args":         ("--service",),
            "metavar":      "service",
            "help":         "(OAuth2)  Service name, if not from the file",
        },
        "handle": {
            "args":         ("--handle",),
            "metavar":      "handle",
            "help":         "(OAuth2)  Handle name, if not from the file",
        },
    }


    def __init__(self, logger, **options):
        self.choices[options['type']](** options)


class Credential(Noun):
    '''
    Operations on credentials.
    '''

    class list(List):
        pass


    class add(Add):
        pass


    class remove(Remove):
        pass


    class listall(ListAll):
        pass


    class get(Get):
        pass


    @classmethod
    def verbs(cls):
        return [cls.list, cls.add, cls.remove, cls.listall, cls.get]



#!/usr/bin/env pytest

# test_dagman_inherit_attrs.py
#
# 

from ornithology import *
import htcondor2
import os

#--------------------------------------------------------------------------------------------
@standup
def condor(test_dir):
    with Condor(local_dir=test_dir / "condor",
                config={
                    "DAGMAN_INHERIT_ATTRS" : "TestAttr1,TestAttr2",
                    "DAGMAN_INHERIT_ATTRS_PREFIX" : "Orig_"
                }) as condor:
        yield condor

#--------------------------------------------------------------------------------------------
@action
def write_dags(path_to_sleep) -> str:
    MAIN_DAG = "test.dag"
    SUB_DAG = "sub.dag"
    with open(MAIN_DAG, "w") as f:
        f.write(f"""
JOB A {{
    executable = {path_to_sleep}
    arguments  = 1
}}
SUBDAG EXTERNAL B {SUB_DAG}
""")

    with open(SUB_DAG, "w") as f:
        f.write(f"""
JOB C {{
    executable = {path_to_sleep}
    arguments  = 1
}}
""")

    return MAIN_DAG

#--------------------------------------------------------------------------------------------
@action
def run_dag(condor, test_dir, write_dags):
    DAG_SUBMIT = htcondor2.Submit().from_dag(write_dags)
    DAG_SUBMIT["My.TestAttr1"] = "1"
    DAG_SUBMIT["My.TestAttr2"] = "true"

    DAG = condor.submit(DAG_SUBMIT)
    assert DAG.wait(condition=ClusterState.all_complete, timeout=60)

    hist_itr = None
    with condor.use_config():
         hist_itr = htcondor2.Schedd().history(constraint="Orig_TestAttr1=!=UNDEFINED && Orig_TestAttr2=!=UNDEFINED",
                                               projection=["DAGNodeName", "Orig_TestAttr1", "Orig_TestAttr2"])
    return hist_itr

#============================================================================================
class TestDAGManInheritAttrs:
    def test_dagman_inherit_attrs(self, run_dag):
        EXPECTED = ["A", "B", "C"]
        for ad in run_dag:
            node_name = ad.get("DAGNodeName")
            assert node_name is not None
            assert node_name in EXPECTED
            EXPECTED.remove(node_name)
            assert ad.get("Orig_TestAttr1") is 1
            assert ad.get("Orig_TestAttr2") is True
        assert len(EXPECTED) == 0




#!/usr/bin/env pytest

import htcondor2

import logging
from getpass import getuser

from ornithology import *


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@action
def the_condor(default_condor):
    return default_condor


@action
def the_held_test_job(the_condor, path_to_sleep, test_dir):
    the_test_job = the_condor.submit(
        {
            "executable":               path_to_sleep,
            "arguments":                1,
            "should_transfer_files":    "yes",
            "transfer_input_files":     "never-exist.file",
            "log":                      (test_dir / "the_test_job.log").as_posix(),
        },
    )

    assert the_test_job.wait(
        condition=ClusterState.all_held,
        timeout=60,
    )

    return the_test_job


class TestFTErrorReporting:

    def test_o_after_i(self, the_held_test_job):
        hold_reasons = the_held_test_job.query(projection=["HoldReason"])
        assert(hold_reasons[0]["HoldReason"].startswith("Transfer input files failure"))
